#
# Copyright 2021-2023 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#


import pytest
from craft_providers.lxd import LXD


@pytest.fixture()
def lxd():
    return LXD()


@pytest.mark.parametrize(
    "sudo",
    [
        pytest.param(False, id="no-sudo"),
        pytest.param(True, marks=pytest.mark.with_sudo, id="with-sudo"),
    ],
)
def test_init(lxd, sudo):
    lxd.init(auto=True, sudo=sudo)


def test_version(lxd):
    version = lxd.version()

    assert isinstance(version, str) is True

    components = version.split(".")

    assert len(components) in [2, 3]


@pytest.mark.parametrize(
    "sudo",
    [
        pytest.param(False, id="no-sudo"),
        pytest.param(True, marks=pytest.mark.with_sudo, id="with-sudo"),
    ],
)
def test_wait_ready(lxd, sudo):
    lxd.wait_ready(sudo=sudo)


def test_is_supported_version(lxd):
    assert lxd.is_supported_version() is True



#
# Copyright 2023 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

import pytest
from craft_providers.bases import get_base_alias, get_base_from_alias, ubuntu
from craft_providers.errors import BaseConfigurationError


def test_get_base_alias():
    assert get_base_alias(("ubuntu", "22.04")) == ubuntu.BuilddBaseAlias.JAMMY


def test_get_base_alias_does_not_exist():
    with pytest.raises(BaseConfigurationError) as exc_info:
        get_base_alias(("ubuntu", "8.04"))

    assert exc_info.value == BaseConfigurationError(
        brief="Base alias not found for BaseName(name='ubuntu', version='8.04')"
    )


def test_get_base_from_alias():
    assert get_base_from_alias(ubuntu.BuilddBaseAlias.JAMMY) == ubuntu.BuilddBase


def test_get_base_from_alias_does_not_exist():
    with pytest.raises(BaseConfigurationError) as exc_info:
        get_base_from_alias("ubuntu 8")  # type: ignore

    assert exc_info.value == BaseConfigurationError(
        brief="Base not found for alias ubuntu 8"
    )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2021-2023 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

"""Manages LXD remotes and provides access to remote images."""

import logging
import warnings
from dataclasses import dataclass
from enum import Enum
from typing import Dict, Union

from craft_providers import Base
from craft_providers.bases import BaseName, almalinux, centos, get_base_alias, ubuntu

from .errors import LXDError
from .lxc import LXC

logger = logging.getLogger(__name__)

BUILDD_RELEASES_REMOTE_NAME = "craft-com.ubuntu.cloud-buildd"
BUILDD_RELEASES_REMOTE_ADDRESS = "https://cloud-images.ubuntu.com/buildd/releases"

BUILDD_DAILY_REMOTE_NAME = "craft-com.ubuntu.cloud-buildd-daily"
BUILDD_DAILY_REMOTE_ADDRESS = "https://cloud-images.ubuntu.com/buildd/daily"

DAILY_REMOTE_NAME = "ubuntu-daily"
DAILY_REMOTE_ADDRESS = "https://cloud-images.ubuntu.com/daily"


class ProtocolType(Enum):
    """Enumeration of protocols for LXD remotes."""

    LXD = "lxd"
    SIMPLESTREAMS = "simplestreams"


@dataclass
class RemoteImage:
    """Contains the name, location, and details of a remote LXD image.

    :param image_name: Name of the image on the remote (e.g. `core22` or `lunar`).
    :param remote_name: Name of the remote server.
    :param remote_address: Address of the remote (can be an IP, FDQN, URL, or token)
    :param remote_protocol: Remote protocol (options are `lxd` and `simplestreams`)
    """

    image_name: str
    remote_name: str
    remote_address: str
    remote_protocol: ProtocolType

    @property
    def is_stable(self) -> bool:
        """Check if the image is stable.

        Images are stable if they are from a release remote.
        Images from daily, devel, and any other remotes are not stable.

        :returns: True if the image is stable.
        """
        if (
            self.remote_name == BUILDD_RELEASES_REMOTE_NAME
            and self.remote_address == BUILDD_RELEASES_REMOTE_ADDRESS
        ):
            # Ubuntu official buildd images
            return True

        if self.remote_name == "images":
            # LXD daily images
            return False

        return False

    def add_remote(self, lxc: LXC) -> None:
        """Add the LXD remote for an image.

        If the remote already exists, it will not be re-added.

        :param lxc: LXC client.
        """
        # TODO verify both the remote name and address
        if self.remote_name in lxc.remote_list():
            logger.debug("Remote %r already exists.", self.remote_name)
        else:
            try:
                lxc.remote_add(
                    remote=self.remote_name,
                    addr=self.remote_address,
                    protocol=self.remote_protocol.value,
                )

            except Exception as exc:
                # the remote adding failed, no matter really how: if it was because a
                # race condition on remote creation (it's not idempotent) and now the
                # remote is there, the purpose of this function is done (otherwise we
                # let the original exception fly)
                if self.remote_name in lxc.remote_list():
                    logger.debug(
                        "Remote %r is present on second check, ignoring exception %r.",
                        self.remote_name,
                        exc,
                    )
                else:
                    raise
            else:
                logger.debug("Remote %r was successfully added.", self.remote_name)


# mapping from supported bases to actual lxd remote images
_PROVIDER_BASE_TO_LXD_REMOTE_IMAGE: Dict[Enum, RemoteImage] = {
    ubuntu.BuilddBaseAlias.BIONIC: RemoteImage(
        image_name="core18",
        remote_name=BUILDD_RELEASES_REMOTE_NAME,
        remote_address=BUILDD_RELEASES_REMOTE_ADDRESS,
        remote_protocol=ProtocolType.SIMPLESTREAMS,
    ),
    ubuntu.BuilddBaseAlias.FOCAL: RemoteImage(
        image_name="core20",
        remote_name=BUILDD_RELEASES_REMOTE_NAME,
        remote_address=BUILDD_RELEASES_REMOTE_ADDRESS,
        remote_protocol=ProtocolType.SIMPLESTREAMS,
    ),
    ubuntu.BuilddBaseAlias.JAMMY: RemoteImage(
        image_name="core22",
        remote_name=BUILDD_RELEASES_REMOTE_NAME,
        remote_address=BUILDD_RELEASES_REMOTE_ADDRESS,
        remote_protocol=ProtocolType.SIMPLESTREAMS,
    ),
    ubuntu.BuilddBaseAlias.NOBLE: RemoteImage(
        image_name="core24",
        remote_name=BUILDD_DAILY_REMOTE_NAME,
        remote_address=BUILDD_DAILY_REMOTE_ADDRESS,
        remote_protocol=ProtocolType.SIMPLESTREAMS,
    ),
    # mantic buildd daily blocked by
    # https://bugs.launchpad.net/cloud-images/+bug/2007419
    ubuntu.BuilddBaseAlias.MANTIC: RemoteImage(
        image_name="mantic",
        remote_name=DAILY_REMOTE_NAME,
        remote_address=DAILY_REMOTE_ADDRESS,
        remote_protocol=ProtocolType.SIMPLESTREAMS,
    ),
    ubuntu.BuilddBaseAlias.DEVEL: RemoteImage(
        image_name="devel",
        remote_name=BUILDD_DAILY_REMOTE_NAME,
        remote_address=BUILDD_DAILY_REMOTE_ADDRESS,
        remote_protocol=ProtocolType.SIMPLESTREAMS,
    ),
    centos.CentOSBaseAlias.SEVEN: RemoteImage(
        image_name="centos/7",
        remote_name="images",
        remote_address="https://images.linuxcontainers.org/images/",
        remote_protocol=ProtocolType.SIMPLESTREAMS,
    ),
    almalinux.AlmaLinuxBaseAlias.NINE: RemoteImage(
        image_name="almalinux/9",
        remote_name="images",
        remote_address="https://images.linuxcontainers.org/images/",
        remote_protocol=ProtocolType.SIMPLESTREAMS,
    ),
}


def get_remote_image(provider_base: Union[Base, str]) -> RemoteImage:
    """Get a RemoteImage for a particular provider base.

    :param provider_base: string containing the provider base

    :returns: the RemoteImage for the provider base
    """
    # temporary backward compatibility before 2.0
    if isinstance(provider_base, str):
        alias = get_base_alias(BaseName("ubuntu", provider_base))
        provider_base = ubuntu.BuilddBase(alias=alias)  # type: ignore

    image = _PROVIDER_BASE_TO_LXD_REMOTE_IMAGE.get(provider_base.alias)
    if not image:
        raise LXDError(
            brief=(
                "could not find a lxd remote image for the provider base "
                f"{provider_base!r}"
            )
        )

    return image


def configure_buildd_image_remote(lxc: LXC = LXC()) -> str:
    """Configure the default buildd image remote.

    This is a deprecated function to maintain the existing API. It will be
    removed with the release of craft-providers 2.0.
    Applications should call `get_remote_image()` and `RemoteImage.add_remote()`.

    :param lxc: LXC client.

    :returns: Name of remote to pass to launcher.
    """
    warnings.warn(
        message=(
            "configure_buildd_image_remote() is deprecated. "
            "Use get_remote_image() and RemoteImage.add_remote()."
        ),
        category=DeprecationWarning,
    )
    # configure the buildd remote for core22
    base = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)
    image = get_remote_image(base)
    image.add_remote(lxc)

    return BUILDD_RELEASES_REMOTE_NAME



#
# Copyright 2021 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

import re
from unittest import mock

import pytest
from craft_providers.multipass import (
    Multipass,
    MultipassError,
    ensure_multipass_is_ready,
)


@pytest.fixture()
def mock_is_installed():
    with mock.patch(
        "craft_providers.multipass._ready.is_installed", return_value=True
    ) as mock_is_installed:
        yield mock_is_installed


@pytest.fixture()
def mock_multipass():
    mock_client = mock.Mock(spec=Multipass)
    mock_client.is_supported_version.return_value = True
    mock_client.version.return_value = "actual-version.0"
    mock_client.minimum_required_version = "min-required-version.0"
    return mock_client


def test_ensure_multipass_is_ready(mock_is_installed, mock_multipass):
    ensure_multipass_is_ready(multipass=mock_multipass)


def test_ensure_multipass_is_ready_fails_incompatible_version(
    mock_is_installed, mock_multipass
):
    mock_multipass.is_supported_version.return_value = False
    match = re.escape(
        "Multipass 'actual-version.0' does not meet the minimum required version "
        "'min-required-version.0'.\nVisit https://multipass.run for instructions "
        "on installing Multipass for your operating system."
    )

    with pytest.raises(MultipassError, match=match):
        ensure_multipass_is_ready(multipass=mock_multipass)


def test_ensure_multipass_is_ready_fails_not_installed(
    mock_is_installed, mock_multipass
):
    mock_is_installed.return_value = False

    match = re.escape(
        "Multipass is required, but not installed.\n"
        "Visit https://multipass.run for instructions "
        "on installing Multipass for your operating system."
    )

    with pytest.raises(MultipassError, match=match):
        ensure_multipass_is_ready(multipass=mock_multipass)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2022-2023 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Multipass Provider class."""

import contextlib
import logging
import pathlib
from dataclasses import dataclass
from enum import Enum
from typing import Dict, Iterator, Optional

from craft_providers import Base, Executor, Provider, base
from craft_providers.bases import ubuntu
from craft_providers.errors import BaseConfigurationError

from ._launch import launch
from ._ready import ensure_multipass_is_ready
from .errors import MultipassError
from .installer import install, is_installed
from .multipass import Multipass
from .multipass_instance import MultipassInstance

logger = logging.getLogger(__name__)


class Remote(Enum):
    """Enumeration of Multipass remotes.

    Multipass uses the name 'snapcraft' for the Ubuntu buildd remote at
    https://cloud-images.ubuntu.com/buildd/.
    """

    RELEASE = "release"
    DAILY = "daily"
    SNAPCRAFT = "snapcraft"


@dataclass
class RemoteImage:
    """Contains the name and details of a remote Multipass image.

    :param remote: Remote server that hosts the image.
    :param image_name: Name of the image on the remote.
    """

    remote: Remote
    image_name: str

    @property
    def is_stable(self) -> bool:
        """Check if the image is stable.

        Images are stable if they are from the snapcraft or release remotes.
        Devel images and images from daily remotes or any other remotes are not stable.

        :returns: True if the image is stable.
        """
        return (
            self.remote in (Remote.RELEASE, Remote.SNAPCRAFT)
            and "devel" not in self.image_name
        )

    @property
    def name(self) -> str:
        """Get the full name of a remote image.

        This name is used to launch an instance with `multipass launch <name>`.

        :returns: Full name of the remote image, formatted as `<remote>:<image_name>`.
        """
        return f"{self.remote.value}:{self.image_name}"


# mapping of Provider bases to Multipass remote images
_BUILD_BASE_TO_MULTIPASS_REMOTE_IMAGE: Dict[Enum, RemoteImage] = {
    ubuntu.BuilddBaseAlias.BIONIC: RemoteImage(
        remote=Remote.SNAPCRAFT, image_name="18.04"
    ),
    ubuntu.BuilddBaseAlias.FOCAL: RemoteImage(
        remote=Remote.SNAPCRAFT, image_name="20.04"
    ),
    ubuntu.BuilddBaseAlias.JAMMY: RemoteImage(
        remote=Remote.SNAPCRAFT, image_name="22.04"
    ),
    ubuntu.BuilddBaseAlias.MANTIC: RemoteImage(
        remote=Remote.RELEASE, image_name="mantic"
    ),
    # this should use `image_name="24.04"` once noble is released (#511)
    ubuntu.BuilddBaseAlias.NOBLE: RemoteImage(
        remote=Remote.SNAPCRAFT, image_name="devel"
    ),
    # devel images are not available on macos
    ubuntu.BuilddBaseAlias.DEVEL: RemoteImage(
        remote=Remote.SNAPCRAFT, image_name="devel"
    ),
}


def _get_remote_image(provider_base: Base) -> RemoteImage:
    """Get a RemoteImage for a particular provider base.

    :param provider_base: String containing the provider base.

    :returns: The RemoteImage for the provider base.

    :raises MultipassError: If the remote image does not exist.
    """
    image = _BUILD_BASE_TO_MULTIPASS_REMOTE_IMAGE.get(provider_base.alias)
    if not image:
        raise MultipassError(
            brief=(
                "could not find a multipass remote image for the provider base "
                f"{provider_base!r}"
            )
        )

    return image


class MultipassProvider(Provider):
    """Multipass build environment provider.

    This class is not stable and is likely to change. This class will be stable and
    recommended for use in the release of craft-providers 2.0.

    :param multipass: Optional Multipass client to use.
    """

    def __init__(self, instance: Multipass = Multipass()) -> None:
        self.multipass = instance

    @property
    def name(self) -> str:
        """Name of the provider."""
        return "Multipass"

    @property
    def install_recommendation(self) -> str:
        """Recommended way to install the provider."""
        return (
            "Visit https://multipass.run/install for instructions to install Multipass."
        )

    @classmethod
    def ensure_provider_is_available(cls) -> None:
        """Ensure provider is available, prompting the user to install it if required.

        :raises MultipassError: if provider is not available.
        """
        if not is_installed():
            install()
        ensure_multipass_is_ready()

    def create_environment(self, *, instance_name: str) -> Executor:
        """Create a bare environment for specified base.

        No initializing, launching, or cleaning up of the environment occurs.

        :param name: Name of the instance.
        """
        return MultipassInstance(name=instance_name)

    @classmethod
    def is_provider_installed(cls) -> bool:
        """Check if provider is installed.

        :returns: True if installed.
        """
        return is_installed()

    @contextlib.contextmanager
    def launched_environment(
        self,
        *,
        project_name: str,
        project_path: pathlib.Path,
        base_configuration: base.Base,
        build_base: Optional[str] = None,
        instance_name: str,
        allow_unstable: bool = False,
    ) -> Iterator[Executor]:
        """Configure and launch environment for specified base.

        When this method loses context, all directories are unmounted and the
        environment is stopped. For more control of environment setup and teardown,
        use `create_environment()` instead.

        :param project_name: Name of the project.
        :param project_path: Path to project.
        :param base_configuration: Base configuration to apply to instance.
        :param build_base: Base to build from. (Deprecated)
        :param instance_name: Name of the instance to launch.
        :param allow_unstable: If true, allow unstable images to be launched.

        :raises MultipassError: If the instance cannot be launched or configured.
        """
        if build_base:
            logger.warning(
                "Deprecated: Parameter 'build_base' is deprecated and should "
                "not be used. The build base now comes from the "
                "base_configuration's alias."
            )

        image = _get_remote_image(base_configuration)

        # only allow launching unstable images when opted-in with `allow_unstable`
        if not image.is_stable and not allow_unstable:
            raise MultipassError(
                brief=f"Cannot launch unstable image {image.name!r}.",
                details=(
                    "Devel or daily images are not guaranteed and are intended for "
                    "experimental use only."
                ),
                resolution=(
                    "Set parameter `allow_unstable` to True to launch unstable images."
                ),
            )
        try:
            instance = launch(
                name=instance_name,
                base_configuration=base_configuration,
                image_name=image.name,
                cpus=2,
                disk_gb=64,
                mem_gb=2,
                auto_clean=True,
            )
        except BaseConfigurationError as error:
            raise MultipassError(str(error)) from error

        try:
            yield instance
        finally:
            # Ensure to unmount everything and stop instance upon completion.
            instance.unmount_all()
            instance.stop()



import sys

sys.path.append("./")
from custom_conf import *

# Configuration file for the Sphinx documentation builder.
# You should not do any modifications to this file. Put your custom
# configuration into the custom_conf.py file.
# If you need to change this file, contribute the changes upstream.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

############################################################
### Extensions
############################################################

extensions = [
    "sphinx_design",
    "sphinx_tabs.tabs",
    "sphinx_reredirects",
    "youtube-links",
    "related-links",
    "custom-rst-roles",
    "terminal-output",
    "sphinx_copybutton",
    "sphinxext.opengraph",
    "myst_parser",
    "sphinxcontrib.jquery",
    "notfound.extension",
]
extensions.extend(custom_extensions)

### Configuration for extensions

# Additional MyST syntax
myst_enable_extensions = ["substitution", "deflist"]

# Used for related links
if "discourse" in html_context:
    html_context["discourse_prefix"] = html_context["discourse"] + "/t/"

# The default for notfound_urls_prefix usually works, but not for
# documentation on documentation.ubuntu.com
if slug:
    notfound_urls_prefix = "/" + slug + "/en/latest/"

notfound_context = {
    "title": "Page not found",
    "body": "<h1>Page not found</h1>\n\n<p>Sorry, but the documentation page that you are looking for was not found.</p>\n<p>Documentation changes over time, and pages are moved around. We try to redirect you to the updated content where possible, but unfortunately, that didn't work this time (maybe because the content you were looking for does not exist in this version of the documentation).</p>\n<p>You can try to use the navigation to locate the content you're looking for, or search for a similar page.</p>\n",
}

# Default image for OGP (to prevent font errors, see
# https://github.com/canonical/sphinx-docs-starter-pack/pull/54 )
if not "ogp_image" in locals():
    ogp_image = "https://assets.ubuntu.com/v1/253da317-image-document-ubuntudocs.svg"

############################################################
### General configuration
############################################################

exclude_patterns = [
    "_build",
    "Thumbs.db",
    ".DS_Store",
    ".sphinx",
]
exclude_patterns.extend(custom_excludes)

rst_epilog = """
.. include:: /reuse/links.txt
"""
if "custom_rst_epilog" in locals():
    rst_epilog = custom_rst_epilog  # type: ignore[name-defined]

source_suffix = {
    ".rst": "restructuredtext",
    ".md": "markdown",
}

if not "conf_py_path" in html_context and "github_folder" in html_context:
    html_context["conf_py_path"] = html_context["github_folder"]

# For ignoring specific links
linkcheck_anchors_ignore_for_url = [r"https://github\.com/.*"]
linkcheck_anchors_ignore_for_url.extend(custom_linkcheck_anchors_ignore_for_url)

############################################################
### Styling
############################################################

# Find the current builder
builder = "dirhtml"
if "-b" in sys.argv:
    builder = sys.argv[sys.argv.index("-b") + 1]

# Setting templates_path for epub makes the build fail
if builder == "dirhtml" or builder == "html":
    templates_path = ["sphinx-docs-starter-pack/.sphinx/_templates"]

# Theme configuration
html_theme = "furo"
html_last_updated_fmt = ""
html_permalinks_icon = "¶"

############################################################
### Additional files
############################################################

html_static_path = ["sphinx-docs-starter-pack/.sphinx/_static"]

html_css_files = [
    "custom.css",
    "header.css",
    "github_issue_links.css",
    "furo_colors.css",
]
html_css_files.extend(custom_html_css_files)

html_js_files = ["header-nav.js"]
if (
    "github_issues" in html_context
    and html_context["github_issues"]
    and not disable_feedback_button
):
    html_js_files.append("github_issue_links.js")
html_js_files.extend(custom_html_js_files)



#
# Copyright 2021 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#


from craft_providers import errors


def test_base_configuration_error():
    error = errors.BaseConfigurationError(brief="foo")

    assert str(error) == "foo"


def test_base_configuration_error_with_details():
    error = errors.BaseConfigurationError(brief="foo", details="bar")

    assert str(error) == "foo\nbar"


def test_base_configuration_error_with_details_resolution():
    error = errors.BaseConfigurationError(
        brief="foo", details="bar", resolution="do this"
    )

    assert str(error) == "foo\nbar\ndo this"


def test_base_compatibility_error():
    error = errors.BaseCompatibilityError(reason="error during foo")

    assert str(error) == (
        "Incompatible base detected: error during foo.\n"
        "Clean incompatible instance and retry the requested operation."
    )


def test_base_compatibility_error_with_details():
    error = errors.BaseCompatibilityError(
        reason="error during foo", details="Some details..."
    )

    assert str(error) == (
        "Incompatible base detected: error during foo.\n"
        "Some details...\n"
        "Clean incompatible instance and retry the requested operation."
    )



#
# Copyright 2021-2023 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

"""Executor module."""

import contextlib
import io
import logging
import pathlib
import subprocess
from abc import ABC, abstractmethod
from typing import Dict, Generator, List, Optional

import craft_providers.util.temp_paths

logger = logging.getLogger(__name__)


class Executor(ABC):
    """Interfaces to execute commands and move data in/out of an environment."""

    @abstractmethod
    def execute_popen(
        self,
        command: List[str],
        *,
        cwd: Optional[pathlib.PurePath] = None,
        env: Optional[Dict[str, Optional[str]]] = None,
        timeout: Optional[float] = None,
        **kwargs,
    ) -> subprocess.Popen:
        """Execute a command in instance, using subprocess.Popen().

        The process' environment will inherit the execution environment's
        default environment (PATH, etc.), but can be additionally configured via
        env parameter.

        :param command: Command to execute.
        :param cwd: Working directory for the process inside the instance.
        :param env: Additional environment to set for process.
        :param timeout: Timeout (in seconds) for the command.
        :param kwargs: Additional keyword arguments to pass.

        :returns: Popen instance.
        """

    @abstractmethod
    def execute_run(
        self,
        command: List[str],
        *,
        cwd: Optional[pathlib.PurePath] = None,
        env: Optional[Dict[str, Optional[str]]] = None,
        timeout: Optional[float] = None,
        check: bool = False,
        **kwargs,
    ) -> subprocess.CompletedProcess:
        """Execute a command using subprocess.run().

        The process' environment will inherit the execution environment's
        default environment (PATH, etc.), but can be additionally configured via
        env parameter.

        :param command: Command to execute.
        :param cwd: Working directory for the process inside the instance.
        :param env: Additional environment to set for process.
        :param timeout: Timeout (in seconds) for the command.
        :param check: Raise an exception if the command fails.
        :param kwargs: Keyword args to pass to subprocess.run().

        :returns: Completed process.

        :raises subprocess.CalledProcessError: if command fails and check is True.
        """

    @abstractmethod
    def pull_file(self, *, source: pathlib.PurePath, destination: pathlib.Path) -> None:
        """Copy a file from the environment to host.

        :param source: Environment file to copy.
        :param destination: Host file path to copy to.  Parent directory
            (destination.parent) must exist.

        :raises FileNotFoundError: If source file or destination's parent
            directory does not exist.
        :raises ProviderError: On error copying file.
        """

    @contextlib.contextmanager
    def temporarily_pull_file(
        self, *, source: pathlib.PurePath, missing_ok: bool = False
    ) -> Generator[Optional[pathlib.Path], None, None]:
        """Copy a file from the environment to a temporary file in the host.

        This is mainly a layer above `pull_file` that pulls the file into a
        temporary path which is cleaned later.

        Works as a context manager, provides the file path in the host as target.

        The temporary file is stored in the home directory where Multipass has access.

        :param source: Environment file to copy.
        :param missing_ok: Do not raise an error if the file does not exist in the
            environment; in this case the target will be None.

        :raises FileNotFoundError: If source file or destination's parent
            directory does not exist (and `missing_ok` is False).
        :raises ProviderError: On error copying file content.
        """
        with craft_providers.util.temp_paths.home_temporary_file() as tmp_file:
            try:
                self.pull_file(source=source, destination=tmp_file)
            except FileNotFoundError:
                if missing_ok:
                    yield None
                else:
                    raise
            else:
                yield tmp_file

    @abstractmethod
    def push_file(self, *, source: pathlib.Path, destination: pathlib.PurePath) -> None:
        """Copy a file from the host into the environment.

        The destination file is overwritten if it exists.

        :param source: Host file to copy.
        :param destination: Target environment file path to copy to.  Parent
            directory (destination.parent) must exist.

        :raises FileNotFoundError: If source file or destination's parent
            directory does not exist.
        :raises ProviderError: On error copying file.
        """

    @abstractmethod
    def push_file_io(
        self,
        *,
        destination: pathlib.PurePath,
        content: io.BytesIO,
        file_mode: str,
        group: str = "root",
        user: str = "root",
    ) -> None:
        """Create or replace a file with specified content and file mode.

        :param destination: Path to file.
        :param content: Contents of file.
        :param file_mode: File mode string (e.g. '0644').
        :param group: File owner group.
        :param user: File owner user.
        """

    @abstractmethod
    def delete(self) -> None:
        """Delete instance."""

    @abstractmethod
    def exists(self) -> bool:
        """Check if instance exists.

        :returns: True if instance exists.
        """

    @abstractmethod
    def mount(self, *, host_source: pathlib.Path, target: pathlib.PurePath) -> None:
        """Mount host source directory to target mount point."""

    @abstractmethod
    def is_running(self) -> bool:
        """Check if instance is running.

        :returns: True if instance is running.
        """



#
# Copyright 2021-2023 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

import pathlib
import subprocess

import pytest
from craft_providers import multipass
from craft_providers.bases import ubuntu


@pytest.fixture()
def simple_file(home_tmp_path):
    """Create a file in the home directory (accessible by Multipass)."""
    file = home_tmp_path / "src.txt"
    file.write_text("this is a test")
    return file


@pytest.mark.smoketest()
def test_smoketest(instance_name, home_tmp_path):
    """Launch an instance and run some basic tasks."""

    assert multipass.is_installed()

    base_configuration = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)

    instance = multipass.launch(
        name=instance_name,
        base_configuration=base_configuration,
        image_name="snapcraft:core22",
        cpus=2,
        disk_gb=8,
        mem_gb=4,
    )

    try:
        assert isinstance(instance, multipass.MultipassInstance)
        assert instance.exists() is True
        assert instance.is_running() is True

        # test execute command
        proc = instance.execute_run(["echo", "hi"], check=True, stdout=subprocess.PIPE)
        assert proc.stdout == b"hi\n"

        # test push file
        test_file = pathlib.Path(home_tmp_path / "src.txt")
        test_file.write_text("this is a test")
        test_file.chmod(0o755)
        destination = pathlib.Path("/tmp/test.txt")
        test_file_pull = pathlib.Path(home_tmp_path / "src2.txt")

        instance.push_file(source=test_file, destination=destination)

        proc = instance.execute_run(
            command=["cat", str(destination)], capture_output=True
        )
        assert proc.stdout.decode() == "this is a test"

        proc = instance.execute_run(
            command=["stat", "--format", "%a:%U:%G", str(destination)],
            capture_output=True,
            text=True,
        )
        assert proc.stdout.strip() == "755:ubuntu:ubuntu"

        # test pull file
        instance.pull_file(source=destination, destination=test_file_pull)
        assert test_file.read_bytes() == test_file_pull.read_bytes()

        # test mount and unmount
        target = pathlib.Path("/tmp/mnt")
        assert instance.is_mounted(host_source=home_tmp_path, target=target) is False

        instance.mount(host_source=home_tmp_path, target=target)
        assert instance.is_mounted(host_source=home_tmp_path, target=target) is True

        proc = instance.execute_run(
            command=["cat", "/tmp/mnt/src.txt"],
            capture_output=True,
        )
        assert proc.stdout == test_file.read_bytes()

        instance.unmount(target=target)
        assert instance.is_mounted(host_source=home_tmp_path, target=target) is False

        # test stop instance
        instance.stop()
        assert instance.is_running() is False
    finally:
        instance.delete()

    # test delete instance
    assert instance.exists() is False



#
# Copyright 2021-2023 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

import subprocess

import pytest
from craft_providers import lxd
from craft_providers.bases import ubuntu

# The LXD tests can be flaky, erroring out with a BaseCompatibilityError:
# "Clean incompatible instance and retry the requested operation."
# This is due to an upstream LXD bug that appears to still be present in LXD 5.14:
# https://github.com/lxc/lxd/issues/11422
pytestmark = pytest.mark.flaky(reruns=3, reruns_delay=2)


# exclude XENIAL because it is not supported for LXD
@pytest.mark.parametrize(
    "alias", set(ubuntu.BuilddBaseAlias) - {ubuntu.BuilddBaseAlias.XENIAL}
)
def test_configure_and_launch_remote(instance_name, alias):
    """Verify remotes are configured and images can be launched."""
    base_configuration = ubuntu.BuilddBase(alias=alias)
    remote_image = lxd.get_remote_image(base_configuration)
    instance = lxd.launch(
        name=instance_name,
        base_configuration=base_configuration,
        image_name=remote_image.image_name,
        image_remote=remote_image.remote_name,
    )

    try:
        assert isinstance(instance, lxd.LXDInstance)
        assert instance.exists() is True
        assert instance.is_running() is True

        proc = instance.execute_run(["echo", "hi"], check=True, stdout=subprocess.PIPE)

        assert proc.stdout == b"hi\n"
    finally:
        instance.delete()


@pytest.mark.parametrize(
    "alias",
    [
        ubuntu.BuilddBaseAlias.BIONIC,
        ubuntu.BuilddBaseAlias.FOCAL,
        ubuntu.BuilddBaseAlias.JAMMY,
    ],
)
def test_configure_and_launch_buildd_remotes(instance_name, alias):
    """Verify function `configure_buildd_image_remote()` can launch core 18|20|22."""
    image_remote = lxd.configure_buildd_image_remote()
    assert image_remote == "craft-com.ubuntu.cloud-buildd"

    base_configuration = ubuntu.BuilddBase(alias=alias)
    instance = lxd.launch(
        name=instance_name,
        base_configuration=base_configuration,
        image_name=alias.value,
        image_remote=image_remote,
    )

    try:
        assert isinstance(instance, lxd.LXDInstance)
        assert instance.exists() is True
        assert instance.is_running() is True

        proc = instance.execute_run(["echo", "hi"], check=True, stdout=subprocess.PIPE)

        assert proc.stdout == b"hi\n"
    finally:
        instance.delete()



import datetime
from typing import Dict, List

# Custom configuration for the Sphinx documentation builder.
# All configuration specific to your project should be done in this file.
#
# The file is included in the common conf.py configuration file.
# You can modify any of the settings below or add any configuration that
# is not covered by the common conf.py file.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

############################################################
### Project information
############################################################

# Product name
project = "Craft Providers"
author = "Canonical Group Ltd"

# Uncomment if your product uses release numbers
# release = '1.0'

# The default value uses the current year as the copyright year.
#
# For static works, it is common to provide the year of first publication.
# Another option is to give the first year and the current year
# for documentation that is often changed, e.g. 2022–2023 (note the en-dash).
#
# A way to check a GitHub repo's creation date is to obtain a classic GitHub
# token with 'repo' permissions here: https://github.com/settings/tokens
# Next, use 'curl' and 'jq' to extract the date from the GitHub API's output:
#
# curl -H 'Authorization: token <TOKEN>' \
#   -H 'Accept: application/vnd.github.v3.raw' \
#   https://api.github.com/repos/canonical/<REPO> | jq '.created_at'

copyright = "2021-%s, %s" % (datetime.date.today().year, author)

## Open Graph configuration - defines what is displayed in the website preview
# The URL of the documentation output
ogp_site_url = "https://canonical-craft-providers.readthedocs-hosted.com/"
# The documentation website name (usually the same as the product name)
ogp_site_name = project
# An image or logo that is used in the preview
ogp_image = "https://assets.ubuntu.com/v1/253da317-image-document-ubuntudocs.svg"

# Update with the favicon for your product (default is the circle of friends)
html_favicon = "sphinx-docs-starter-pack/.sphinx/_static/favicon.png"

# (Some settings must be part of the html_context dictionary, while others
#  are on root level. Don't move the settings.)
html_context = {
    # Change to the link to your product website (without "https://")
    "product_page": "github.com/canonical/craft-providers",
    # Add your product tag to ".sphinx/_static" and change the path
    # here (start with "_static"), default is the circle of friends
    "product_tag": "_static/tag.png",
    # Change to the discourse instance you want to be able to link to
    # using the :discourse: metadata at the top of a file
    # (use an empty value if you don't want to link)
    "discourse": "",
    # Change to the GitHub info for your project
    "github_url": "https://github.com/canonical/craft-providers",
    # Change to the branch for this version of the documentation
    "github_version": "main",
    # Change to the folder that contains the documentation
    # (usually "/" or "/docs/")
    "github_folder": "/docs/",
    # Change to an empty value if your GitHub repo doesn't have issues enabled.
    # This will disable the feedback button and the issue link in the footer.
    "github_issues": "enabled",
    # Controls the existence of Previous / Next buttons at the bottom of pages
    # Valid options: none, prev, next, both
    "sequential_nav": "none",
}

# If your project is on documentation.ubuntu.com, specify the project
# slug (for example, "lxd") here.
slug = ""

############################################################
### Redirects
############################################################

# Set up redirects (https://documatt.gitlab.io/sphinx-reredirects/usage.html)
# For example: 'explanation/old-name.html': '../how-to/prettify.html',

redirects: Dict[str, str] = {}

############################################################
### Link checker exceptions
############################################################

# Links to ignore when checking links

linkcheck_ignore: List[str] = ["http://127.0.0.1:8000"]

# Pages on which to ignore anchors
# (This list will be appended to linkcheck_anchors_ignore_for_url)

custom_linkcheck_anchors_ignore_for_url: List[str] = []

############################################################
### Additions to default configuration
############################################################

## The following settings are appended to the default configuration.
## Use them to extend the default functionality.

# Add extensions
custom_extensions: List[str] = [
    "sphinx.ext.autodoc",
]

# Add files or directories that should be excluded from processing.
custom_excludes: List[str] = [
    "sphinx-docs-starter-pack/*",
]

# Add CSS files (located in .sphinx/_static/)
custom_html_css_files: List[str] = []

# Add JavaScript files (located in .sphinx/_static/)
custom_html_js_files: List[str] = []

## The following settings override the default configuration.

# Specify a reST string that is included at the end of each file.
# If commented out, use the default (which pulls the reuse/links.txt
# file into each reST file).
# custom_rst_epilog = ''

# By default, the documentation includes a feedback button at the top.
# You can disable it by setting the following configuration to True.
disable_feedback_button = False

############################################################
### Additional configuration
############################################################

## Add any configuration that is not covered by the common conf.py file.



#
# Copyright 2021-2023 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

import hashlib
import io
import os
import pathlib
import shutil
import subprocess
import sys
import tempfile
from unittest import mock
from unittest.mock import call

import pytest
from craft_providers import errors
from craft_providers.lxd import LXC, LXDError, LXDInstance
from logassert import Exact  # type: ignore

# These names include invalid characters so a lxd-compatible instance_name
# is generated. This ensures an Instance's `name` and `instance_name` are
# differentiated when testing.
_TEST_INSTANCE = {
    "name": "test-instance-$",
    "instance-name": "test-instance-fa2d407652a1c51f6019",
}

_STOPPED_INSTANCE = {
    "name": "stopped-instance-$",
    "instance-name": "stopped-instance-b4598eebe37eb50c4612",
}

_INVALID_INSTANCE = {
    "name": "invalid-instance-$",
    "instance-name": "invalid-instance-86be3150e96a80a04e31",
}


@pytest.fixture()
def project_path(tmp_path):
    project_path = tmp_path / "git" / "project"
    project_path.mkdir(parents=True)
    return project_path


@pytest.fixture(autouse=True)
def mock_lxc(project_path):
    with mock.patch("craft_providers.lxd.lxd_instance.LXC", spec=LXC) as lxc:
        lxc.list.return_value = [
            {"name": _TEST_INSTANCE["instance-name"], "status": "Running"},
            {"name": _STOPPED_INSTANCE["instance-name"], "status": "Stopped"},
        ]
        lxc.config_device_show.return_value = {
            "test_mount": {
                "path": "/root/project",
                "source": project_path.as_posix(),
                "type": "disk",
            },
            "disk-/target": {
                "path": "/target",
                "source": "/source",
                "type": "disk",
            },
        }
        lxc.info.return_value = {
            "environment": {"kernel_features": {}},
        }
        yield lxc


@pytest.fixture()
def mock_named_temporary_file():
    with mock.patch(
        "craft_providers.lxd.lxd_instance.tempfile.NamedTemporaryFile",
        spec=tempfile.NamedTemporaryFile,
    ) as mock_tf:
        mock_tf.return_value.__enter__.return_value.name = "test-tmp-file"
        yield mock_tf.return_value


@pytest.fixture()
def mock_shutil_copyfileobj():
    with mock.patch.object(shutil, "copyfileobj") as mock_copyfileobj:
        yield mock_copyfileobj


@pytest.fixture()
def mock_os_unlink():
    with mock.patch.object(os, "unlink") as mock_unlink:
        yield mock_unlink


@pytest.fixture()
def instance(mock_lxc):
    return LXDInstance(name=_TEST_INSTANCE["name"], lxc=mock_lxc)


def test_config_get(mock_lxc, instance):
    instance.config_get(key="test-key")

    assert mock_lxc.mock_calls == [
        call.config_get(
            instance_name="test-instance-fa2d407652a1c51f6019",
            key="test-key",
            project="default",
            remote="local",
        )
    ]


def test_config_set(mock_lxc, instance):
    instance.config_set(key="test-key", value="test-value")

    assert mock_lxc.mock_calls == [
        call.config_set(
            instance_name="test-instance-fa2d407652a1c51f6019",
            key="test-key",
            value="test-value",
            project="default",
            remote="local",
        )
    ]


def test_info(mock_lxc, instance):
    instance.info()

    assert mock_lxc.mock_calls == [
        call.info(
            instance_name="test-instance-fa2d407652a1c51f6019",
            project="default",
            remote="local",
        )
    ]


def test_push_file_io(
    mock_lxc,
    mock_named_temporary_file,
    mock_shutil_copyfileobj,
    mock_os_unlink,
    instance,
):
    mock_lxc.exec.side_effect = [
        None,
        None,
        None,
        None,
    ]

    instance.push_file_io(
        destination=pathlib.PurePosixPath("/etc/test.conf"),
        content=io.BytesIO(b"foo"),
        file_mode="0644",
    )

    assert mock_lxc.mock_calls == [
        mock.call.file_push(
            instance_name=instance.instance_name,
            source=pathlib.Path("test-tmp-file"),
            destination=pathlib.PurePosixPath("/etc/test.conf"),
            mode="0644",
            project=instance.project,
            remote=instance.remote,
        ),
        mock.call.exec(
            instance_name=instance.instance_name,
            command=[
                "chown",
                "root:root",
                "/etc/test.conf",
            ],
            cwd=None,
            project=instance.project,
            remote=instance.remote,
            runner=subprocess.run,
            capture_output=True,
            check=True,
            timeout=60,
        ),
    ]

    assert mock_named_temporary_file.mock_calls == [
        mock.call.__enter__(),
        mock.call.__enter__().flush(),
        mock.call.__exit__(None, None, None),
    ]
    assert mock_shutil_copyfileobj.mock_calls == [
        mock.call(mock.ANY, mock_named_temporary_file.__enter__.return_value)
    ]


def test_push_file_io_error(mock_lxc, instance):
    error = subprocess.CalledProcessError(
        -1, ["chown", "root:root", "/etc/test.conf"], "test stdout", "test stderr"
    )

    mock_lxc.exec.side_effect = error

    with pytest.raises(LXDError) as exc_info:
        instance.push_file_io(
            destination=pathlib.PurePosixPath("/etc/test.conf"),
            content=io.BytesIO(b"foo"),
            file_mode="0644",
        )

    assert exc_info.value == LXDError(
        brief=(
            "Failed to create file '/etc/test.conf' "
            f"in instance '{_TEST_INSTANCE['instance-name']}'."
        ),
        details=errors.details_from_called_process_error(error),
    )


def test_delete(mock_lxc, instance):
    instance.delete()

    assert mock_lxc.mock_calls == [
        mock.call.delete(
            instance_name=instance.instance_name,
            project=instance.project,
            remote=instance.remote,
            force=True,
        )
    ]


def test_execute_popen(mock_lxc, instance):
    instance.execute_popen(command=["test-command", "flags"], input="foo")

    assert mock_lxc.mock_calls == [
        mock.call.exec(
            instance_name=instance.instance_name,
            command=["test-command", "flags"],
            cwd=None,
            project=instance.project,
            remote=instance.remote,
            runner=subprocess.Popen,
            input="foo",
            timeout=None,
        )
    ]


def test_execute_popen_with_cwd(mock_lxc, instance):
    instance.execute_popen(
        command=["test-command", "flags"],
        cwd=pathlib.PurePosixPath("/tmp"),
        input="foo",
    )

    assert mock_lxc.mock_calls == [
        mock.call.exec(
            instance_name=instance.instance_name,
            command=["test-command", "flags"],
            cwd="/tmp",
            project=instance.project,
            remote=instance.remote,
            runner=subprocess.Popen,
            input="foo",
            timeout=None,
        )
    ]


def test_execute_popen_with_env(mock_lxc, instance):
    instance.execute_popen(command=["test-command", "flags"], env={"foo": "bar"})

    assert mock_lxc.mock_calls == [
        mock.call.exec(
            instance_name=instance.instance_name,
            command=["env", "foo=bar", "test-command", "flags"],
            cwd=None,
            project=instance.project,
            remote=instance.remote,
            runner=subprocess.Popen,
            timeout=None,
        )
    ]


def test_execute_run(mock_lxc, instance):
    instance.execute_run(command=["test-command", "flags"], input="foo")

    assert mock_lxc.mock_calls == [
        mock.call.exec(
            instance_name=instance.instance_name,
            command=["test-command", "flags"],
            cwd=None,
            project=instance.project,
            remote=instance.remote,
            runner=subprocess.run,
            input="foo",
            timeout=None,
            check=False,
        )
    ]


def test_execute_run_with_cwd(mock_lxc, instance):
    instance.execute_run(
        command=["test-command", "flags"],
        cwd=pathlib.PurePosixPath("/tmp"),
        input="foo",
    )

    assert mock_lxc.mock_calls == [
        mock.call.exec(
            instance_name=instance.instance_name,
            command=["test-command", "flags"],
            cwd="/tmp",
            project=instance.project,
            remote=instance.remote,
            runner=subprocess.run,
            input="foo",
            timeout=None,
            check=False,
        )
    ]


@pytest.mark.parametrize("check", [True, False])
def test_execute_run_with_check(check, mock_lxc, instance):
    instance.execute_run(command=["test-command", "flags"], input="foo", check=check)

    assert mock_lxc.mock_calls == [
        mock.call.exec(
            instance_name=instance.instance_name,
            command=["test-command", "flags"],
            cwd=None,
            project=instance.project,
            remote=instance.remote,
            runner=subprocess.run,
            input="foo",
            timeout=None,
            check=check,
        )
    ]


def test_execute_run_with_default_command_env(mock_lxc):
    instance = LXDInstance(
        name=_TEST_INSTANCE["name"],
        default_command_environment={"env_key": "some-value"},
        lxc=mock_lxc,
    )

    instance.execute_run(command=["test-command", "flags"], env={"foo": "bar"})

    assert mock_lxc.mock_calls == [
        mock.call.exec(
            instance_name=instance.instance_name,
            command=["env", "env_key=some-value", "foo=bar", "test-command", "flags"],
            cwd=None,
            project=instance.project,
            remote=instance.remote,
            runner=subprocess.run,
            timeout=None,
            check=False,
        )
    ]


def test_execute_run_with_default_command_env_unset(mock_lxc):
    instance = LXDInstance(
        name=_TEST_INSTANCE["name"],
        default_command_environment={"env_key": "some-value"},
        lxc=mock_lxc,
    )

    instance.execute_run(
        command=["test-command", "flags"], env={"foo": "bar", "env_key": None}
    )

    assert mock_lxc.mock_calls == [
        mock.call.exec(
            instance_name=instance.instance_name,
            command=["env", "-u", "env_key", "foo=bar", "test-command", "flags"],
            cwd=None,
            project=instance.project,
            remote=instance.remote,
            runner=subprocess.run,
            timeout=None,
            check=False,
        )
    ]


def test_execute_run_with_env(mock_lxc, instance):
    instance.execute_run(command=["test-command", "flags"], env={"foo": "bar"})

    assert mock_lxc.mock_calls == [
        mock.call.exec(
            instance_name=instance.instance_name,
            command=["env", "foo=bar", "test-command", "flags"],
            cwd=None,
            project=instance.project,
            remote=instance.remote,
            runner=subprocess.run,
            timeout=None,
            check=False,
        )
    ]


def test_execute_run_with_env_unset(mock_lxc, instance):
    instance.execute_run(
        command=["test-command", "flags"], env={"foo": "bar", "TERM": None}
    )

    assert mock_lxc.mock_calls == [
        mock.call.exec(
            instance_name=instance.instance_name,
            command=[
                "env",
                "foo=bar",
                "-u",
                "TERM",
                "test-command",
                "flags",
            ],
            cwd=None,
            project=instance.project,
            remote=instance.remote,
            runner=subprocess.run,
            timeout=None,
            check=False,
        )
    ]


def test_exists(mock_lxc, instance):
    assert instance.exists() is True
    assert mock_lxc.mock_calls == [
        mock.call.list(project=instance.project, remote=instance.remote)
    ]


def test_exists_false(mock_lxc):
    instance = LXDInstance(name="does-not-exist", lxc=mock_lxc)

    assert instance.exists() is False
    assert mock_lxc.mock_calls == [
        mock.call.list(project=instance.project, remote=instance.remote)
    ]


def test_get_disk_devices_path_parse_error(mock_lxc, instance):
    mock_lxc.config_device_show.return_value = {
        "mount_missing_path": {
            "source": "/foo",
            "type": "disk",
        },
    }

    with pytest.raises(LXDError) as exc_info:
        instance._get_disk_devices()

    assert exc_info.value == LXDError(
        brief="Failed to parse lxc device 'mount_missing_path'.",
        details=(
            "* Device configuration:"
            " {'mount_missing_path': {'source': '/foo', 'type': 'disk'}}"
        ),
    )


def test_get_disk_devices_source_parse_error(mock_lxc, instance):
    mock_lxc.config_device_show.return_value = {
        "mount_missing_source": {
            "path": "/foo",
            "type": "disk",
        },
    }

    with pytest.raises(LXDError) as exc_info:
        instance._get_disk_devices()

    assert exc_info.value == LXDError(
        brief="Failed to parse lxc device 'mount_missing_source'.",
        details=(
            "* Device configuration:"
            " {'mount_missing_source': {'path': '/foo', 'type': 'disk'}}"
        ),
    )


def test_is_mounted_false(mock_lxc, instance):
    project_path = pathlib.Path.home() / "not-mounted"

    assert (
        instance.is_mounted(
            host_source=pathlib.Path(project_path),
            target=pathlib.PurePosixPath("/root/project"),
        )
        is False
    )

    assert mock_lxc.mock_calls == [
        mock.call.config_device_show(
            instance_name=instance.instance_name,
            project=instance.project,
            remote=instance.remote,
        )
    ]


def test_is_mounted_true(mock_lxc, instance, project_path):
    assert (
        instance.is_mounted(
            host_source=project_path,
            target=pathlib.PurePosixPath("/root/project"),
        )
        is True
    )

    assert mock_lxc.mock_calls == [
        mock.call.config_device_show(
            instance_name=instance.instance_name,
            project=instance.project,
            remote=instance.remote,
        )
    ]


def test_is_running_false(mock_lxc):
    instance = LXDInstance(name=_STOPPED_INSTANCE["name"], lxc=mock_lxc)

    assert instance.is_running() is False

    assert mock_lxc.mock_calls == [
        mock.call.list(project=instance.project, remote=instance.remote)
    ]


def test_is_running_true(mock_lxc, instance):
    assert instance.is_running() is True

    assert mock_lxc.mock_calls == [
        mock.call.list(project=instance.project, remote=instance.remote)
    ]


def test_is_running_error(mock_lxc):
    instance = LXDInstance(name=_INVALID_INSTANCE["name"], lxc=mock_lxc)

    with pytest.raises(LXDError) as exc_info:
        instance.is_running()

    assert exc_info.value == LXDError(
        brief=f"Instance '{_INVALID_INSTANCE['instance-name']}' does not exist.",
    )


def test_launch(mock_lxc, instance):
    instance.launch(
        image="22.04",
        image_remote="ubuntu",
    )

    assert mock_lxc.mock_calls == [
        mock.call.info(project=instance.project, remote=instance.remote),
        mock.call.launch(
            config_keys={},
            ephemeral=False,
            instance_name=instance.instance_name,
            image="22.04",
            image_remote="ubuntu",
            project=instance.project,
            remote=instance.remote,
        ),
    ]


@pytest.mark.skipif(sys.platform == "win32", reason="unsupported on windows")
def test_launch_all_opts(mock_lxc, instance):
    instance.launch(
        image="22.04",
        image_remote="ubuntu",
        ephemeral=True,
        map_user_uid=True,
    )

    uid = str(os.getuid())
    assert mock_lxc.mock_calls == [
        mock.call.info(project=instance.project, remote=instance.remote),
        mock.call.launch(
            config_keys={"raw.idmap": f"both {uid} 0"},
            ephemeral=True,
            instance_name=instance.instance_name,
            image="22.04",
            image_remote="ubuntu",
            project=instance.project,
            remote=instance.remote,
        ),
    ]


def test_launch_with_mknod(mock_lxc, instance):
    mock_lxc.info.return_value = {
        "environment": {"kernel_features": {"seccomp_listener": "true"}}
    }

    instance.launch(
        image="22.04",
        image_remote="ubuntu",
    )

    assert mock_lxc.mock_calls == [
        mock.call.info(project=instance.project, remote=instance.remote),
        mock.call.launch(
            config_keys={
                "security.syscalls.intercept.mknod": "true",
            },
            ephemeral=False,
            instance_name=instance.instance_name,
            image="22.04",
            image_remote="ubuntu",
            project=instance.project,
            remote=instance.remote,
        ),
    ]


def test_mount(mock_lxc, tmp_path, instance):
    """Verify calls to mount a directory."""
    instance.mount(host_source=tmp_path, target=pathlib.PurePosixPath("/mnt/foo"))

    assert mock_lxc.mock_calls == [
        mock.call.config_device_show(
            instance_name=instance.instance_name,
            project=instance.project,
            remote=instance.remote,
        ),
        mock.call.config_device_add_disk(
            instance_name=instance.instance_name,
            source=tmp_path,
            path=pathlib.PurePosixPath("/mnt/foo"),
            device="disk-/mnt/foo",
            project=instance.project,
            remote=instance.remote,
        ),
    ]


def test_mount_already_mounted(mock_lxc, instance, project_path):
    """Do not mount if directory is already mounted."""
    instance.mount(
        host_source=project_path, target=pathlib.PurePosixPath("/root/project")
    )

    assert mock_lxc.mock_calls == [
        mock.call.config_device_show(
            instance_name=instance.instance_name,
            project=instance.project,
            remote=instance.remote,
        )
    ]


def test_pull_file(mock_lxc, instance, tmp_path):
    mock_lxc.exec.return_value = mock.Mock(returncode=0)

    source = pathlib.PurePosixPath("/tmp/src.txt")
    destination = tmp_path / "dst.txt"

    instance.pull_file(
        source=source,
        destination=destination,
    )

    assert mock_lxc.mock_calls == [
        mock.call.exec(
            instance_name=instance.instance_name,
            command=["test", "-f", "/tmp/src.txt"],
            cwd=None,
            project=instance.project,
            remote=instance.remote,
            runner=subprocess.run,
            check=False,
            timeout=60,
        ),
        mock.call.file_pull(
            instance_name=instance.instance_name,
            source=source,
            destination=destination,
            project=instance.project,
            remote=instance.remote,
        ),
    ]


def test_pull_file_no_source(mock_lxc, instance, tmp_path):
    mock_lxc.exec.return_value = mock.Mock(returncode=1)

    source = pathlib.PurePosixPath("/tmp/src.txt")
    destination = tmp_path / "dst.txt"

    with pytest.raises(FileNotFoundError) as exc_info:
        instance.pull_file(
            source=source,
            destination=destination,
        )

    assert mock_lxc.mock_calls == [
        mock.call.exec(
            instance_name=instance.instance_name,
            command=["test", "-f", "/tmp/src.txt"],
            cwd=None,
            project=instance.project,
            remote=instance.remote,
            runner=subprocess.run,
            check=False,
            timeout=60,
        ),
    ]
    assert str(exc_info.value) == "File not found: '/tmp/src.txt'"


def test_pull_file_no_parent_directory(mock_lxc, instance, tmp_path):
    mock_lxc.exec.return_value = mock.Mock(returncode=0)

    source = pathlib.PurePosixPath("/tmp/src.txt")
    destination = tmp_path / "not-created" / "dst.txt"

    with pytest.raises(FileNotFoundError) as exc_info:
        instance.pull_file(
            source=source,
            destination=destination,
        )

    assert mock_lxc.mock_calls == [
        mock.call.exec(
            instance_name=instance.instance_name,
            command=["test", "-f", "/tmp/src.txt"],
            cwd=None,
            project=instance.project,
            remote=instance.remote,
            runner=subprocess.run,
            check=False,
            timeout=60,
        ),
    ]
    assert str(exc_info.value) == f"Directory not found: {str(destination.parent)!r}"


def test_push_file(mock_lxc, instance, tmp_path):
    mock_lxc.exec.return_value = mock.Mock(returncode=0)

    source = tmp_path / "src.txt"
    source.write_text("this is a test")
    destination = pathlib.PurePosixPath("/tmp/dst.txt")

    instance.push_file(
        source=source,
        destination=destination,
    )

    assert mock_lxc.mock_calls == [
        mock.call.exec(
            instance_name=instance.instance_name,
            command=["test", "-d", "/tmp"],
            cwd=None,
            project=instance.project,
            remote=instance.remote,
            runner=subprocess.run,
            check=False,
            timeout=60,
        ),
        mock.call.file_push(
            instance_name=instance.instance_name,
            source=source,
            destination=destination,
            project=instance.project,
            remote=instance.remote,
            gid=0,
            uid=0,
        ),
    ]


def test_push_file_no_source(mock_lxc, instance, tmp_path):
    source = tmp_path / "src.txt"
    destination = pathlib.PurePosixPath("/tmp/dst.txt")

    with pytest.raises(FileNotFoundError) as exc_info:
        instance.push_file(
            source=source,
            destination=destination,
        )

    assert mock_lxc.mock_calls == []
    assert str(exc_info.value) == f"File not found: {str(source)!r}"


def test_push_file_no_parent_directory(mock_lxc, instance, tmp_path):
    mock_lxc.exec.return_value = mock.Mock(returncode=1)

    source = tmp_path / "src.txt"
    source.write_text("this is a test")
    destination = pathlib.PurePosixPath("/tmp/dst.txt")

    with pytest.raises(FileNotFoundError) as exc_info:
        instance.push_file(
            source=source,
            destination=destination,
        )

    assert mock_lxc.mock_calls == [
        mock.call.exec(
            instance_name=instance.instance_name,
            command=["test", "-d", "/tmp"],
            cwd=None,
            project=instance.project,
            remote=instance.remote,
            runner=subprocess.run,
            check=False,
            timeout=60,
        ),
    ]
    assert str(exc_info.value) == "Directory not found: '/tmp'"


def test_start(mock_lxc, instance):
    instance.start()

    assert mock_lxc.mock_calls == [
        mock.call.start(
            instance_name=instance.instance_name,
            project=instance.project,
            remote=instance.remote,
        )
    ]


def test_restart(mock_lxc, instance):
    instance.restart()

    assert mock_lxc.mock_calls == [
        call.restart(
            instance_name="test-instance-fa2d407652a1c51f6019",
            project="default",
            remote="local",
        )
    ]


def test_stop(mock_lxc, instance):
    instance.stop()

    assert mock_lxc.mock_calls == [
        mock.call.stop(
            instance_name=instance.instance_name,
            project=instance.project,
            remote=instance.remote,
        )
    ]


def test_supports_mount(instance):
    instance.remote = "local"

    assert instance.supports_mount() is True

    instance.remote = "some-remote"

    assert instance.supports_mount() is False


def test_unmount(mock_lxc, instance):
    instance.unmount(target=pathlib.PurePosixPath("/root/project"))

    assert mock_lxc.mock_calls == [
        mock.call.config_device_show(
            instance_name=instance.instance_name,
            project=instance.project,
            remote=instance.remote,
        ),
        mock.call.config_device_remove(
            instance_name=instance.instance_name,
            device="test_mount",
            project=instance.project,
            remote=instance.remote,
        ),
    ]


def test_unmount_all(mock_lxc, instance):
    instance.unmount_all()

    assert mock_lxc.mock_calls == [
        mock.call.config_device_show(
            instance_name=instance.instance_name,
            project=instance.project,
            remote=instance.remote,
        ),
        mock.call.config_device_remove(
            instance_name=instance.instance_name,
            device="test_mount",
            project=instance.project,
            remote=instance.remote,
        ),
        mock.call.config_device_remove(
            instance_name=instance.instance_name,
            device="disk-/target",
            project=instance.project,
            remote=instance.remote,
        ),
    ]


def test_unmount_error(mock_lxc, instance):
    mock_lxc.config_device_show.return_value = {
        "disk-/target": {
            "path": "/target",
            "source": "/source",
            "type": "disk",
        },
    }

    with pytest.raises(LXDError) as exc_info:
        instance.unmount(target=pathlib.PurePosixPath("not-mounted"))

    assert exc_info.value == LXDError(
        brief=(
            "Failed to unmount 'not-mounted'"
            f" in instance '{_TEST_INSTANCE['instance-name']}' - no such disk."
        ),
        details=(
            "* Disk device configuration: {'disk-/target':"
            " {'path': '/target', 'source': '/source', 'type': 'disk'}}"
        ),
    )


@pytest.mark.parametrize(
    "name",
    [
        "t",
        "test",
        "test1",
        "test-1",
        "this-is-40-characters-xxxxxxxxxxxxxxxxxx",
        "this-is-63-characters-xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx",
    ],
)
def test_set_instance_name_unchanged(logs, mock_lxc, name):
    """Verify names that are already compliant are not changed."""
    instance = LXDInstance(
        name=name,
        lxc=mock_lxc,
    )

    assert instance.instance_name == name
    assert Exact(f"Set LXD instance name to {name!r}") in logs.debug


@pytest.mark.parametrize(
    ("name", "expected_name"),
    [
        # trim away invalid beginning characters
        ("1test", "test"),
        ("123test", "test"),
        ("-test", "test"),
        ("1-2-3-test", "test"),
        # trim away invalid ending characters
        ("test-", "test"),
        ("test--", "test"),
        ("test1-", "test1"),
        # trim away invalid characters
        ("test$", "test"),
        ("test-!@#$%^&*()test", "test-test"),
        ("$1test", "test"),
        ("test-$", "test"),
        # this name contains invalid characters so it gets converted, even
        # though it is 63 characters
        (
            "this-is-63-characters-with-invalid-characters-$$$xxxxxxxxxxxxxX",
            "this-is-63-characters-with-invalid-chara",
        ),
        # this name is longer than 63 characters, so it gets converted
        (
            "this-is-70-characters-and-valid-xxxxxxxXxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx",
            "this-is-70-characters-and-valid-xxxxxxxX",
        ),
    ],
)
def test_set_instance_name(logs, mock_lxc, name, expected_name):
    """Verify name is compliant with LXD naming conventions."""
    instance = LXDInstance(
        name=name,
        lxc=mock_lxc,
    )

    # compute hash
    hashed_name = hashlib.sha1(name.encode()).hexdigest()[:20]

    assert instance.instance_name == f"{expected_name}-{hashed_name}"
    assert len(instance.instance_name) <= 63
    assert Exact(f"Set LXD instance name to {instance.instance_name!r}") in logs.debug


def test_set_instance_name_hash_value(mock_lxc):
    """Verify hash is formatted as expected.

    The first 20 characters of the SHA-1 hash of
    "hello-world$" is 'b993dc52118c0f489570'

    The name "hello-world$" should be hashed, not the trimmed name "hello-world".
    """
    instance = LXDInstance(
        name="hello-world$",
        lxc=mock_lxc,
    )

    assert instance.instance_name == "hello-world-b993dc52118c0f489570"
    assert len(instance.instance_name) <= 63


@pytest.mark.parametrize(
    "name",
    [
        "",
        "-",
        "$$$",
        "-$-$-",
    ],
)
def test_set_instance_name_invalid(mock_lxc, name):
    """Verify invalid names raise an error."""
    with pytest.raises(LXDError) as error:
        LXDInstance(name=name, lxc=mock_lxc)

    assert error.value == LXDError(
        brief=f"failed to create LXD instance with name {name!r}.",
        details="name must contain at least one alphanumeric character",
    )



#
# Copyright 2021 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

"""Project helper utilities."""
import logging

from .lxc import LXC

logger = logging.getLogger(__name__)


def create_with_default_profile(
    *,
    lxc: LXC,
    project: str,
    profile: str = "default",
    profile_project: str = "default",
    remote: str = "local",
) -> None:
    """Create a project with a valid default profile.

    LXD does not set a valid profile on newly created projects.  This will
    create a project and set the profile to match the specified profile,
    typically the default.

    :param project: Name of project to create.
    :param remote: Name of remote.
    :param profile_name: Name of profile to copy.

    :raises LXDError: on unexpected error.
    """
    lxc.project_create(project=project, remote=remote)

    # Retrieve config from specified (default) project profile.
    config = lxc.profile_show(profile=profile, project=profile_project, remote=remote)

    # Set config.
    lxc.profile_edit(profile="default", project=project, config=config, remote=remote)


def purge(*, lxc: LXC, project: str, remote: str = "local") -> None:
    """Purge a project including its instances and images.

    The lxc command does not provide a straight-forward option to purge a
    project.  This helper will purge anything related to a specified one.

    :param project: Name of project to delete.
    :param remote: Name of remote.

    :raises LXDError: on unexpected error.
    """
    logger.debug("Purging project %r on remote %r.", project, remote)
    projects = lxc.project_list(remote=remote)
    if project not in projects:
        logger.debug(
            "Attempted to purge non-existent project %r on remote %r.", project, remote
        )
        return

    # Cleanup any outstanding instance_names.
    for instance_name in lxc.list_names(project=project, remote=remote):
        logger.debug(
            "Deleting instance %r from project %r on remote %r.",
            instance_name,
            project,
            remote,
        )
        lxc.delete(
            instance_name=instance_name,
            project=project,
            remote=remote,
            force=True,
        )

    # Cleanup any outstanding images.
    for image in lxc.image_list(project=project):
        logger.debug(
            "Deleting image %r from project %r on remote %r.", image, project, remote
        )
        lxc.image_delete(image=image["fingerprint"], project=project, remote=remote)

    # Cleanup project.
    logger.debug("Deleting project %r on remote %r.", project, remote)
    lxc.project_delete(project=project, remote=remote)

    logger.debug("Project %r on remote %r was purged successfully.", project, remote)



#
# Copyright 2021-2023 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

import io
import pathlib
import subprocess

import pytest
from craft_providers.lxd import LXDInstance

from . import conftest

pytestmark = [
    # These tests are flaky on very busy systems.
    # https://github.com/lxc/lxd/issues/11422
    # https://github.com/lxc/lxd/issues/11890
    pytest.mark.flaky(reruns=2, reruns_delay=1),
]


@pytest.fixture()
def instance(instance_name, project):
    with conftest.tmp_instance(
        name=instance_name,
        project=project,
    ):
        instance = LXDInstance(name=instance_name, project=project)

        yield instance


@pytest.fixture(
    scope="module",
    params=[
        pytest.param(version, marks=pytest.mark.xdist_group(name=version))
        for version in ["18.04", "20.04", "22.04", "23.04"]
    ],
)
def reusable_instance(reusable_instance_name, request):
    """Reusable instance for tests that don't require a fresh instance."""
    name = f"{reusable_instance_name}-reusable-{request.param}"
    with conftest.tmp_instance(
        name=name,
        image=request.param,
        ephemeral=False,
        project="default",
    ):
        instance = LXDInstance(name=name, project="default")

        yield instance


@pytest.mark.parametrize("content", [b"", b"\x00\xaa\xbb\xcc", b"test-string"])
@pytest.mark.parametrize("mode", ["644", "600", "755"])
@pytest.mark.parametrize(("user", "group"), [("root", "root"), ("ubuntu", "ubuntu")])
def test_push_file_io(reusable_instance, content, mode, user, group):
    try:
        reusable_instance.push_file_io(
            destination=pathlib.PurePosixPath("/tmp/file-test.txt"),
            content=io.BytesIO(content),
            file_mode=mode,
            user=user,
            group=group,
        )

        proc = reusable_instance.execute_run(
            command=["cat", "/tmp/file-test.txt"],
            capture_output=True,
        )

        assert proc.stdout == content

        proc = reusable_instance.execute_run(
            command=["stat", "--format", "%a:%U:%G", "/tmp/file-test.txt"],
            capture_output=True,
            text=True,
        )

        assert proc.stdout.strip() == f"{mode}:{user}:{group}"
    finally:
        reusable_instance.execute_run(
            command=["rm", "-f", "/tmp/file-test.txt"],
            capture_output=True,
        )


def test_delete(instance):
    assert instance.exists() is True

    instance.delete()

    assert instance.exists() is False


def test_execute_popen(reusable_instance):
    with reusable_instance.execute_popen(
        command=["pwd"],
        stdout=subprocess.PIPE,
        text=True,
    ) as proc:
        stdout, _ = proc.communicate()

    assert stdout.strip() == "/root"


def test_execute_popen_cwd(reusable_instance):
    with reusable_instance.execute_popen(
        command=["pwd"],
        cwd=pathlib.PurePosixPath("/tmp"),
        stdout=subprocess.PIPE,
        text=True,
    ) as proc:
        stdout, _ = proc.communicate()

    assert stdout.strip() == "/tmp"


def test_execute_run(reusable_instance):
    proc = reusable_instance.execute_run(
        command=["pwd"],
        capture_output=True,
        text=True,
    )

    assert proc.stdout.strip() == "/root"


def test_execute_run_cwd(reusable_instance):
    proc = reusable_instance.execute_run(
        command=["pwd"],
        cwd=pathlib.PurePosixPath("/"),
        capture_output=True,
        text=True,
    )

    assert proc.stdout.strip() == "/"


def test_exists(reusable_instance):
    assert reusable_instance.exists() is True


def test_exists_false():
    fake_instance = LXDInstance(name="does-not-exist")

    assert fake_instance.exists() is False


def test_launch(instance_name):
    instance = LXDInstance(name=instance_name)

    assert instance.exists() is False

    instance.launch(
        image="20.04",
        image_remote="ubuntu",
    )

    try:
        assert instance.exists() is True
    finally:
        instance.delete()


@pytest.mark.parametrize(
    "name",
    [
        "test-name",
        "more-than-63-characters-xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx",
        "more-than-63-characters-and-invalid-characters-$$$xxxxxxxxxxxxxxxxxxxx",
        "invalid-characters-$$$",
    ],
)
def test_launch_with_name(instance_name, name):
    """Verify we can launch an instance even when we pass in an invalid name."""
    # prepend the tester's random instance name
    name = f"{instance_name}-{name}"
    instance = LXDInstance(name=name)

    assert instance.exists() is False

    instance.launch(
        image="20.04",
        image_remote="ubuntu",
    )

    try:
        assert instance.exists() is True
    finally:
        instance.delete()


def test_mount_unmount(reusable_instance, tmp_path):
    tmp_path.chmod(0o755)

    host_source = tmp_path
    target = pathlib.PurePosixPath("/tmp/mnt")

    test_file = host_source / "test.txt"
    test_file.write_text("this is a test")

    assert reusable_instance.is_mounted(host_source=host_source, target=target) is False

    reusable_instance.mount(host_source=host_source, target=target)

    assert reusable_instance.is_mounted(host_source=host_source, target=target) is True

    proc = reusable_instance.execute_run(
        command=["cat", "/tmp/mnt/test.txt"],
        capture_output=True,
    )

    assert proc.stdout == test_file.read_bytes()

    reusable_instance.unmount(target=target)

    assert reusable_instance.is_mounted(host_source=host_source, target=target) is False


def test_mount_unmount_all(reusable_instance, tmp_path):
    tmp_path.chmod(0o755)

    source1 = tmp_path / "1"
    source1.mkdir()
    target_1 = pathlib.PurePosixPath("/tmp/mnt/1")

    source2 = tmp_path / "2"
    source2.mkdir()
    target_2 = pathlib.PurePosixPath("/tmp/mnt/2")

    reusable_instance.mount(host_source=source1, target=target_1)
    reusable_instance.mount(host_source=source2, target=target_2)

    assert reusable_instance.is_mounted(host_source=source1, target=target_1) is True
    assert reusable_instance.is_mounted(host_source=source2, target=target_2) is True

    reusable_instance.unmount_all()

    assert reusable_instance.is_mounted(host_source=source1, target=target_1) is False
    assert reusable_instance.is_mounted(host_source=source2, target=target_2) is False


def test_start_stop(reusable_instance):
    assert reusable_instance.is_running() is True

    reusable_instance.stop()

    assert reusable_instance.is_running() is False

    reusable_instance.start()

    assert reusable_instance.is_running() is True


def test_info(reusable_instance):
    info = reusable_instance.info()

    assert info.get("Name") == reusable_instance.instance_name



#
# Copyright 2021-2023 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

"""LXD Instance Provider."""

import logging
import os
import re
import subprocess
import sys
import threading
import time
from datetime import datetime, timedelta, timezone
from pathlib import Path
from typing import Optional

from craft_providers import Base, ProviderError, bases
from craft_providers.errors import details_from_called_process_error

from .errors import LXDError
from .lxc import LXC
from .lxd_instance import LXDInstance
from .lxd_instance_status import ProviderInstanceStatus
from .project import create_with_default_profile

logger = logging.getLogger(__name__)


class InstanceTimer(threading.Thread):
    """Timer for update instance that still alive."""

    __active: bool = True
    __interval: float
    __instance: LXDInstance

    def __init__(self, instance: LXDInstance, interval: int = 3) -> None:
        """Initialize the timer.

        :param instance: LXD instance to update.
        :param interval: Interval in seconds to update the instance timer.
        """
        self.__instance = instance
        self.__interval = interval
        super().__init__(daemon=True)

    def run(self) -> None:
        """Run the timer."""
        while self.__active:
            now = datetime.now(timezone.utc).isoformat()
            try:
                self.__instance.config_set("user.craft_providers.timer", now)
                logger.debug("Set instance timer to %r", now)
            except LXDError:
                # Error in timer update is not critical
                logging.exception("Error updating instance timer")
            time.sleep(self.__interval)

        try:
            self.__instance.config_set("user.craft_providers.timer", "DONE")
            logger.debug("Set instance timer to 'DONE'")
        except LXDError:
            # Error in timer update is not critical
            logging.exception("Error updating instance timer")
        logger.debug("Instance timer update stopped.")

    def stop(self) -> None:
        """Stop the timer."""
        self.__active = False


def _create_instance(
    *,
    instance: LXDInstance,
    base_instance: Optional[LXDInstance],
    base_configuration: Base,
    image_name: str,
    image_remote: str,
    ephemeral: bool,
    map_user_uid: bool,
    uid: Optional[int],
    project: str,
    remote: str,
    lxc: LXC,
) -> None:
    """Launch and setup an instance from an image.

    If a base instance is passed, copy the instance to the base instance.

    Preconditions: The LXD project exists and the instance and base instance (if
    defined) do not exist.

    :param instance: LXD instance to launch and setup
    :param base_instance: LXD instance to be created as a copy of the instance. If the
    base_instance is None, a base instance will not be created.
    :param base_configuration: Base configuration to apply to instance.
    :param image_name: LXD image to use, e.g. "20.04".
    :param image_remote: LXD image to use, e.g. "ubuntu".
    :param ephemeral: After the instance is stopped, delete it.
    :param map_user_uid: Map host uid/gid to instance's root uid/gid.
    :param uid: The uid to be mapped, if ``map_user_uid`` is enabled.
    :param project: LXD project to create instance in.
    :param remote: LXD remote to create instance on.
    :param lxc: LXC client.
    """
    logger.info("Creating new instance from remote")

    # Lockable base instance creation. Only one caller can create the base instance.
    # Other callers will will get LXDError and wait until the base instance is created.
    if base_instance:
        logger.info("Creating new base instance from remote")
        logger.debug(
            "Creating new base instance from image %r from remote %r",
            image_name,
            image_remote,
        )
        base_instance.launch(
            image=image_name,
            image_remote=image_remote,
            ephemeral=False,  # base instance should not ephemeral
            map_user_uid=map_user_uid,
            uid=uid,
        )
        base_instance_status = base_instance.config_get("user.craft_providers.status")

        # Skip the base configuration if the instance is already configured.
        if base_instance_status != ProviderInstanceStatus.FINISHED.value:
            logger.debug("Setting up base instance %r", base_instance.instance_name)
            base_instance.config_set(
                "user.craft_providers.status", ProviderInstanceStatus.PREPARING.value
            )
            config_timer = InstanceTimer(base_instance)
            config_timer.start()

            # The base configuration shouldn't mount cache directories because if
            # they get deleted, copying the base instance will fail.
            base_configuration.setup(executor=base_instance, mount_cache=False)
            _set_timezone(
                base_instance,
                base_instance.project,
                base_instance.remote,
                base_instance.lxc,
            )
            # set the full instance name as image description
            base_instance.config_set("image.description", base_instance.name)
            base_instance.config_set(
                "user.craft_providers.status", ProviderInstanceStatus.FINISHED.value
            )
            config_timer.stop()
            base_instance.stop()

        # Copy the base instance to the instance.
        logger.info("Creating new instance from base instance")
        logger.debug(
            "Creating new instance %r from base instance %r",
            instance.instance_name,
            base_instance.instance_name,
        )
        lxc.copy(
            source_remote=base_instance.remote,
            source_instance_name=base_instance.instance_name,
            destination_remote=remote,
            destination_instance_name=instance.instance_name,
            project=project,
        )
        _set_timezone(instance, project, remote, instance.lxc)
    else:
        logger.debug(
            "Creating new instance from image %r from remote %r",
            image_name,
            image_remote,
        )
        instance.launch(
            image=image_name,
            image_remote=image_remote,
            ephemeral=ephemeral,
            map_user_uid=map_user_uid,
            uid=uid,
        )
        instance_status = instance.config_get("user.craft_providers.status")

        # Skip the base configuration if the instance is already configured.
        if instance_status != ProviderInstanceStatus.FINISHED.value:
            logger.info("Setting up instance")
            logger.debug("Setting up instance %r", instance.instance_name)
            instance.config_set(
                "user.craft_providers.status", ProviderInstanceStatus.PREPARING.value
            )
            config_timer = InstanceTimer(instance)
            config_timer.start()
            base_configuration.setup(executor=instance)
            _set_timezone(instance, project, remote, instance.lxc)
            instance.config_set(
                "user.craft_providers.status", ProviderInstanceStatus.FINISHED.value
            )
            config_timer.stop()
            if not ephemeral:
                # stop ephemeral instances will delete them immediately
                instance.stop()

    # after creating the base instance, the id map can be set
    if map_user_uid:
        _set_id_map(
            instance=instance, lxc=instance.lxc, project=project, remote=remote, uid=uid
        )

    # now restart and wait for the instance to be ready
    if ephemeral:
        # ephemeral instances can only be restarted
        instance.restart()
    else:
        instance.start()
    # Warmup the instance to ensure we have everything set up.
    # There are cases where setup won't do everything to the base instance that
    # warmup does to the final instance (such as mounting cache directories).
    base_configuration.warmup(executor=instance)


def _ensure_project_exists(
    *, create: bool, project: str, remote: str, lxc: LXC
) -> None:
    """Check if project exists and create it if it does not exist.

    :param create: Create project if not found.
    :param project: LXD project name to create.
    :param remote: LXD remote to create project on.
    :param lxc: LXC client.

    :raises LXDError: on error.
    """
    projects = lxc.project_list(remote)
    if project in projects:
        return

    if create:
        create_with_default_profile(project=project, remote=remote, lxc=lxc)
    else:
        raise LXDError(
            brief=f"LXD project {project!r} not found on remote {remote!r}.",
            details=f"Available projects: {projects!r}",
        )


def _formulate_base_instance_name(
    *, image_name: str, image_remote: str, compatibility_tag: str
) -> str:
    """Compute the base instance name.

    :param image_remote: Name of source image's remote (e.g. ubuntu).
    :param image_name: Name of source image (e.g. 20.04). The image name should include
    the architecture to ensure uniqueness amongst multiple architectures built on the
    same platform.
    :param compatibility_tag: Compatibility tag of base configuration applied to the
    base instance.

    :returns: Name of (compatible) base instance.
    """
    return "-".join(["base-instance", compatibility_tag, image_remote, image_name])


def _is_valid(*, instance: LXDInstance, expiration: timedelta) -> bool:
    """Check if an instance is valid.

    Instances are valid if they are ready and not expired (too old).
    An instance is ready if it is set up and stopped.
    An instance's age is measured by its creation date. For example, if the expiration
    is 90 days, then the instance will expire 91 days after it was created.

    If errors occur during the validity check, the instance is assumed to be invalid.

    :param instance: LXD instance to check the validity of.
    :param expiration: How long an instance will be valid from its creation date.

    :returns: True if the instance is valid. False otherwise.
    """
    logger.debug("Checking validity of instance %r.", instance.instance_name)

    # capture instance info
    try:
        info = instance.info()
    except LXDError as raised:
        # if the base instance info can't be retrieved, consider it invalid
        logger.debug("Could not get instance info with error: %s", raised)
        return False

    creation_date_raw = info.get("Created")

    # if the base instance does not have a creation date, consider it invalid
    if not creation_date_raw:
        logger.debug("Instance does not have a 'Created' date.")
        return False

    # parse datetime
    try:
        creation_date = datetime.strptime(creation_date_raw, "%Y/%m/%d %H:%M %Z")
    except ValueError as raised:
        # if the date can't be parsed, consider it invalid
        logger.debug("Could not parse instance's 'Created' date with error: %r", raised)
        return False

    expiration_date = datetime.now() - expiration
    if creation_date < expiration_date:
        logger.debug(
            "Instance is expired (Instance creation date: %s, expiration date: %s).",
            creation_date,
            expiration_date,
        )
        return False

    try:
        _wait_for_instance_ready(instance)
    except LXDError as err:
        logger.debug("Instance is not valid: %s", err)
        return False

    logger.debug("Instance is valid.")
    return True


def _launch_existing_instance(
    *,
    instance: LXDInstance,
    lxc: LXC,
    project: str,
    remote: str,
    auto_clean: bool,
    base_configuration: Base,
    ephemeral: bool,
    map_user_uid: bool,
    uid: Optional[int],
) -> bool:
    """Start and warmup an existing instance.

    Autocleaning allows for an incompatible instance to be deleted and rebuilt rather
    than raising an error. Autocleaning will be triggered if:
    - the base instance is incompatible for any reason (e.g. wrong OS)
    - the instance's existing id map does not match the id map passed to this function

    :param instance: LXD instance to launch
    :param lxc: LXC client.
    :param project: LXD project of the instance.
    :param remote: LXD remote of the instance.
    :param auto_clean: If true, clean incompatible instances.
    :param base_configuration: Base configuration to apply to the instance.
    :param ephemeral: If the instance is ephemeral, it will not be launched.
    Instead, the instance will be deleted and the function will return false.
    :param map_user_uid: True if host uid/gid should be mapped to the instance's
    root uid/gid.
    :param uid: The host uid that should be mapped, if ``map_user_uid`` is enabled.
    If none, use the current user's uid.

    :returns: True if the instance was started and warmed up. False otherwise.

    :raises BaseCompatibilityError: If the instance is incompatible.
    """
    if not _check_id_map(
        instance=instance,
        lxc=lxc,
        project=project,
        remote=remote,
        map_user_uid=map_user_uid,
        uid=uid,
    ):
        if auto_clean:
            logger.debug(
                "Cleaning incompatible instance %r (reason: %s).",
                instance.instance_name,
                "the instance's id map ('raw.idmap') is not configured as expected",
            )
            # delete the instance so a new instance can be created
            instance.delete()
            return False
        raise bases.BaseCompatibilityError(
            reason="the instance's id map ('raw.idmap') is not configured as expected"
        )

    if ephemeral:
        logger.debug("Instance exists and is ephemeral. Cleaning instance.")
        instance.delete()
        return False

    if instance.is_running():
        logger.debug("Instance exists and is running.")
    else:
        logger.debug("Instance exists and is not running. Starting instance.")
        instance.start()

    try:
        base_configuration.warmup(executor=instance)
    except bases.BaseCompatibilityError as error:
        # delete the instance so a new instance can be created
        if auto_clean:
            logger.debug(
                "Cleaning incompatible instance %r (reason: %s).",
                instance.instance_name,
                error.reason,
            )
            instance.delete()
            return False
        raise

    return True


def _check_id_map(
    *,
    instance: LXDInstance,
    lxc: LXC,
    project: str,
    remote: str,
    map_user_uid: bool,
    uid: Optional[int],
) -> bool:
    """Check if the instance's id map matches the uid passed as an argument.

    :param instance: LXD instance to set the idmap of
    :param lxc: LXC client.
    :param project: LXD project to create instance in.
    :param remote: LXD remote to create instance on.
    :param map_user_uid: If true, check that the instance has an id map.
    If false, check that the instance has no id map.
    :param uid: The uid that should already be mapped. If None, the current user's uid
    is used.

    :returns: True if the instance's id map matches the expected id map.
    """
    if uid is None:
        uid = os.getuid()

    configured_id_map = lxc.config_get(
        instance_name=instance.instance_name,
        key="raw.idmap",
        project=project,
        remote=remote,
    )

    # if the id map is not configured and should not be configured, then return True
    if not configured_id_map and not map_user_uid:
        return True

    match = re.fullmatch("both ([0-9]+) 0", configured_id_map)

    # if the id map is not exactly what craft-providers configured, then return False
    if not match:
        logger.debug(
            "Unexpected id map for %r (expected 'both %s 0', got %r).",
            instance.instance_name,
            uid,
            configured_id_map,
        )
        return False

    # get the configured uid from the id map
    configured_uid = int(match.group(1))

    return configured_uid == uid


def _set_id_map(
    *,
    instance: LXDInstance,
    lxc: LXC = LXC(),
    project: str = "default",
    remote: str = "local",
    uid: Optional[int] = None,
) -> None:
    """Configure the instance's id map.

    By default, LXD creates unprivileged containers by using user namespaces to map
    privileged uid/gids in the instance to unprivileged uid/gids in the host.
    In order to mount a directory from the host in an LXD instance, the host user's ids
    must be mapped before the instance is started.

    The instance needs to be stopped or restarted for the id map to take effect.

    :param instance: LXD instance to set the idmap of
    :param lxc: LXC client.
    :param project: LXD project to create instance in.
    :param remote: LXD remote to create instance on.
    :param uid: The uid to be mapped. If not supplied, the current user's uid is used.
    """
    configured_id_map = ""
    if uid is None:
        uid = os.getuid()

    lxc.config_set(
        instance_name=instance.instance_name,
        key="raw.idmap",
        value=f"both {uid!s} 0",
        project=project,
        remote=remote,
    )

    # There was a race in LXD that lxc config returns success,
    # but the config was not applied. See https://github.com/canonical/lxd/issues/12189
    configured_id_map = lxc.config_get(
        instance_name=instance.instance_name,
        key="raw.idmap",
        project=project,
        remote=remote,
    )
    configured_id_map_list = [line.split() for line in configured_id_map.splitlines()]

    logger.debug(
        "Got LXD idmap for instance %r: %r", instance.instance_name, configured_id_map
    )

    if ["both", str(uid), "0"] not in configured_id_map_list:
        raise LXDError(
            brief=f"Failed to set idmap for instance {instance.instance_name!r}",
            details=f"Expected idmap: 'both {uid!s} 0', got {configured_id_map!r}",
        )


def _set_timezone(instance: LXDInstance, project: str, remote: str, lxc: LXC) -> None:
    """Set the instance's timezone to match the host's timezone.

    Sets the instance timezone only on Linux hosts due to usage of the
    `timedatectl` utility.
    """
    # no-op on non-linux platforms because `timedatectl` won't exist
    if sys.platform != "linux":
        logger.debug("Not setting timezone because host is not Linux.")
        return

    try:
        timezone = subprocess.run(
            ["timedatectl", "show", "-p", "Timezone", "--value"],
            capture_output=True,
            check=True,
            text=True,
        ).stdout.strip()
    except subprocess.CalledProcessError as error:
        # `timedatectl` support on the host is not an explicit requirement, so just
        # log the failure and no-op
        logger.debug(
            "Not setting instance's timezone because host timezone could not "
            f"be determined: {details_from_called_process_error(error)}"
        )
        return

    logger.debug(f"Setting instance timezone to match host timezone {timezone!r}.")

    # set the timezone via the TZ environment variable
    lxc.config_set(
        instance_name=instance.instance_name,
        key="environment.TZ",
        value=timezone,
        project=project,
        remote=remote,
    )


def _wait_for_instance_ready(instance: LXDInstance) -> None:
    """Wait for an instance to be ready.

    If another process created an instance and is still running, then wait for
    the other process to finish setting up the instance.

    :param instance: LXD instance to wait for.

    :raises LXDError:
    - If the instance is not ready and the process that created the
    instance is no longer running.
    - If the function times out while waiting for another process
    to set up the instance.
    - On any failure to get config data or info from the instance.
    """
    instance_state = instance.info().get("Status")
    instance_status = instance.config_get("user.craft_providers.status")

    # check if the instance is ready
    if (
        instance_status == ProviderInstanceStatus.FINISHED.value
        and instance_state == "STOPPED"
    ):
        logger.debug("Instance %r is ready.", instance.instance_name)
        return

    logger.debug(
        "Instance %r is not ready (State: %s, Status: %s)",
        instance.instance_name,
        instance_state,
        instance_status,
    )

    # LXD is linux-only, but verify the platform anyways
    if sys.platform == "linux":
        # get the PID of the process that created the instance
        pid = instance.config_get("user.craft_providers.pid")

        # an empty string means there was no value set
        if pid == "":
            raise LXDError(
                brief="Instance is not ready.",
                details=(
                    f"Instance {instance.instance_name!r} is not ready and does not "
                    "have the pid of the process that created the instance."
                ),
            )

        # check if the PID is active
        if Path(f"/proc/{pid}").exists():
            logger.debug(
                "Process that created the instance %r is still running (pid %s).",
                instance.instance_name,
                pid,
            )
        else:
            raise LXDError(
                brief="Instance is not ready.",
                details=(
                    f"Instance {instance.instance_name!r} is not ready and the process "
                    f"(pid {pid}) that created the instance is inactive."
                ),
            )
    else:
        logger.debug("Skipping PID check because system is not linux")

    # if the PID is active, wait for the instance to be ready
    instance.lxc.check_instance_status(
        instance_name=instance.instance_name,
        project=instance.project,
        remote=instance.remote,
    )


def launch(
    name: str,
    *,
    base_configuration: Base,
    image_name: str,
    image_remote: str,
    auto_clean: bool = False,
    auto_create_project: bool = False,
    ephemeral: bool = False,
    map_user_uid: bool = False,
    uid: Optional[int] = None,
    use_snapshots: Optional[bool] = None,
    use_base_instance: bool = False,
    project: str = "default",
    remote: str = "local",
    lxc: LXC = LXC(),
    expiration: timedelta = timedelta(days=90),
) -> LXDInstance:
    """Create, start, and configure an instance.

    On the first run of an application, an instance will be launched from an image
    (i.e. an image from  https://cloud-images.ubuntu.com). The instance is setup
    according to the Base configuration passed to this function.

    After setup, a copy of this instance is saved (or cached) as a 'base instance'.
    This is done to reduce setup time on subsequent runs. When the application requests
    a new instance on a subsequent run, the base instance will be copied to create the
    new instance. This instance is run through a small subset of the setup, which is
    referred to as 'warmup'.

    To keep build environments clean, consistent, and up-to-date, any base instance
    older than 3 months (90 days) is deleted and recreated. This 90 day default can be
    changed with the `expiration` parameter.

    :param name: Name of instance.
    :param base_configuration: Base configuration to apply to the instance.
    :param image_name: LXD image to use, e.g. "20.04".
    :param image_remote: LXD image to use, e.g. "ubuntu".
    :param auto_clean: If true and the existing instance is incompatible, then the
    instance will be deleted and rebuilt. If false and the existing instance is
    incompatible, then a BaseCompatibilityError is raised.
    :param auto_create_project: Automatically create LXD project, if needed.
    :param ephemeral: After the instance is stopped, delete it. Non-ephemeral instances
    cannot be converted to ephemeral instances, so if the instance already exists, it
    will be deleted, then recreated as an ephemeral instance.
    :param map_user_uid: Map host uid/gid to instance's root uid/gid.
    :param uid: The uid to be mapped, if ``map_user_uid`` is enabled.
    :param use_base_instance: Use the base instance mechanisms to reduce setup time.
    :param use_snapshots: Deprecated parameter replaced by `use_base_instance`.
    :param project: LXD project to create instance in.
    :param remote: LXD remote to create instance on.
    :param lxc: LXC client.
    :param expiration: How long a base instance will be valid from its creation date.

    :returns: LXD instance.

    :raises BaseConfigurationError: on unexpected error configuration base.
    :raises BaseCompatibilityError: if instance is incompatible with the base.
    :raises LXDError: on unexpected LXD error.
    :raises ProviderError: if name of instance collides with base instance name.
    """
    # TODO: create a private class to reduce the parameters passed between methods

    if use_snapshots:
        logger.warning(
            "Deprecated: Parameter 'use_snapshots' is deprecated. "
            "Use parameter 'use_base_instance' instead."
        )
        use_base_instance = use_snapshots

    _ensure_project_exists(
        create=auto_create_project, project=project, remote=remote, lxc=lxc
    )
    instance = LXDInstance(
        name=name,
        project=project,
        remote=remote,
        default_command_environment=base_configuration.get_command_environment(),
    )
    logger.debug(
        "Checking for instance %r in project %r in remote %r",
        instance.instance_name,
        project,
        remote,
    )

    # if the existing instance could not be launched, then continue on so a new
    # instance can be created (this can occur when `auto_clean` triggers the
    # instance to be deleted or if the instance is supposed to be ephemeral)
    if instance.exists() and _launch_existing_instance(
        instance=instance,
        lxc=lxc,
        project=project,
        remote=remote,
        auto_clean=auto_clean,
        base_configuration=base_configuration,
        ephemeral=ephemeral,
        map_user_uid=map_user_uid,
        uid=uid,
    ):
        return instance

    logger.debug("Instance %r does not exist.", instance.instance_name)

    if not use_base_instance:
        logger.debug("Using base instances is disabled.")
        _create_instance(
            instance=instance,
            base_instance=None,
            base_configuration=base_configuration,
            image_name=image_name,
            image_remote=image_remote,
            ephemeral=ephemeral,
            map_user_uid=map_user_uid,
            uid=uid,
            project=project,
            remote=remote,
            lxc=lxc,
        )
        return instance

    base_instance_name = _formulate_base_instance_name(
        image_name=image_name,
        image_remote=image_remote,
        compatibility_tag=base_configuration.compatibility_tag,
    )
    base_instance = LXDInstance(
        name=base_instance_name,
        project=project,
        remote=remote,
        default_command_environment=base_configuration.get_command_environment(),
    )
    logger.debug(
        "Checking for base instance %r in project %r in remote %r",
        base_instance.instance_name,
        project,
        remote,
    )

    # an application could formulate an instance name that matches the base instance's
    # name, which would break calls to `lxc.copy()`
    if instance.instance_name == base_instance.instance_name:
        raise ProviderError(
            brief="instance name cannot match the base instance name: "
            f"{instance.instance_name!r}",
            resolution="change name of instance",
        )

    # the base instance does not exist, so create a new instance and base instance
    if not base_instance.exists():
        logger.debug("Base instance %r does not exist.", base_instance.instance_name)
        _create_instance(
            instance=instance,
            base_instance=base_instance,
            base_configuration=base_configuration,
            image_name=image_name,
            image_remote=image_remote,
            ephemeral=ephemeral,
            map_user_uid=map_user_uid,
            uid=uid,
            project=project,
            remote=remote,
            lxc=lxc,
        )
        return instance

    # the base instance exists but is not valid, so delete it then create a new
    # instance and base instance
    if not _is_valid(instance=base_instance, expiration=expiration):
        logger.debug(
            "Base instance %r is not valid. Deleting base instance.",
            base_instance.instance_name,
        )
        base_instance.delete()
        _create_instance(
            instance=instance,
            base_instance=base_instance,
            base_configuration=base_configuration,
            image_name=image_name,
            image_remote=image_remote,
            ephemeral=ephemeral,
            map_user_uid=map_user_uid,
            uid=uid,
            project=project,
            remote=remote,
            lxc=lxc,
        )
        return instance

    # at this point, there is a valid base instance to be copied to a new instance
    logger.info("Creating instance from base instance")
    logger.debug(
        "Creating instance from base instance %r.", base_instance.instance_name
    )

    # the base instance is not expected to be running but check for safety
    if base_instance.is_running():
        logger.debug("Stopping base instance.")

    lxc.copy(
        source_remote=remote,
        source_instance_name=base_instance.instance_name,
        destination_remote=remote,
        destination_instance_name=instance.instance_name,
        project=project,
    )

    # the newly copied instance should not be running, but check anyways
    if instance.is_running():
        logger.debug("Instance is already running.")
        instance.stop()

    # set the id map while the instance is not running
    if map_user_uid:
        _set_id_map(instance=instance, lxc=lxc, project=project, remote=remote, uid=uid)

    # instance is now ready to be started and warmed up
    instance.start()

    # change the hostname from the base instance's hostname
    base_configuration._setup_hostname(executor=instance)

    base_configuration.warmup(executor=instance)

    # set timezone on warmup because the host's timezone may have changed since the
    # instance was setup
    _set_timezone(instance, project, remote, lxc)

    return instance



#
# Copyright 2021-2023 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

"""Base configuration module."""


import enum
import io
import logging
import math
import os
import pathlib
import re
import subprocess
import sys
from abc import ABC, abstractmethod
from enum import Enum
from textwrap import dedent
from typing import Dict, List, Optional, Type, final

from pydantic import ValidationError

from craft_providers.actions import snap_installer
from craft_providers.actions.snap_installer import Snap, SnapInstallationError
from craft_providers.const import (
    RETRY_WAIT,
    TIMEOUT_COMPLEX,
    TIMEOUT_SIMPLE,
    TIMEOUT_UNPREDICTABLE,
)
from craft_providers.errors import (
    BaseCompatibilityError,
    BaseConfigurationError,
    NetworkError,
    details_from_called_process_error,
)
from craft_providers.executor import Executor
from craft_providers.instance_config import InstanceConfiguration
from craft_providers.util import retry
from craft_providers.util.os_release import parse_os_release

logger = logging.getLogger(__name__)


class Base(ABC):
    """Interface for providers to configure instantiated environments.

    Defines how to setup/configure an environment that has been instantiated by
    a provider and prepare it for some operation, e.g. execute build.  It must
    account for:

    (1) the OS type and version.

    (2) the provided image that was launched, e.g. bootstrapping a minimal image
    versus a more fully featured one.

    (3) any dependencies that are required for the operation to complete, e.g.
    installed applications, networking configuration, etc.  This includes any
    environment configuration that the application will assume is available.

    :cvar compatibility_tag: Tag/Version for variant of build configuration and
        setup.  Any change to this version would indicate that prior [versioned]
        instances are incompatible and must be cleaned.  As such, any new value
        should be unique to old values (e.g. incrementing).  It is suggested to
        extend this tag, not overwrite it, e.g.: compatibility_tag =
        f"{appname}-{Base.compatibility_tag}.{apprevision}" to ensure base
        compatibility levels are maintained.

    :param cache_path: (Optional) Path to the shared cache directory. If this is
        provided, shared cache directories will be mounted as appropriate. Some
        directories depend on the base implementation.
    """

    _environment: Dict[str, Optional[str]]
    _hostname: str
    _instance_config_path = pathlib.PurePosixPath("/etc/craft-instance.conf")
    _instance_config_class: Type[InstanceConfiguration] = InstanceConfiguration
    _snaps: Optional[List[Snap]] = None
    _packages: Optional[List[str]] = None
    _retry_wait: float = RETRY_WAIT
    _timeout_simple: Optional[float] = TIMEOUT_SIMPLE
    _timeout_complex: Optional[float] = TIMEOUT_COMPLEX
    _timeout_unpredictable: Optional[float] = TIMEOUT_UNPREDICTABLE
    _cache_path: Optional[pathlib.Path] = None
    alias: Enum
    compatibility_tag: str = "base-v7"

    @abstractmethod
    def __init__(
        self,
        *,
        alias: enum.Enum,
        compatibility_tag: Optional[str] = None,
        environment: Optional[Dict[str, Optional[str]]] = None,
        hostname: str = "craft-instance",
        snaps: Optional[List] = None,
        packages: Optional[List[str]] = None,
        use_default_packages: bool = True,
        cache_path: Optional[pathlib.Path] = None,
    ) -> None:
        pass

    @staticmethod
    def default_command_environment() -> Dict[str, Optional[str]]:
        """Provide default command environment dictionary.

        The minimum environment for the image to be configured and function
        properly.  This contains the default environment for most Linux systems
        in /etc/environment.

        :returns: Dictionary of environment key/values.
        """
        return {
            "PATH": "/usr/local/sbin:/usr/local/bin:"
            "/usr/sbin:/usr/bin:/sbin:/bin:/snap/bin"
        }

    def _set_hostname(self, hostname: str) -> None:
        """Set hostname.

        hostname naming convention:
        - between 1 and 63 characters long
        - be made up exclusively of letters, numbers, and hyphens from the ASCII table
        - not begin or end with a hyphen

        If needed, the provided hostname will be trimmed to meet naming conventions.

        :param hostname: hostname to set
        :raises BaseConfigurationError: if the hostname contains no
          alphanumeric characters
        """
        # truncate to 63 characters
        truncated_name = hostname[:63]

        # remove anything that is not an alphanumeric character or hyphen
        name_with_valid_chars = re.sub(r"[^\w-]", "", truncated_name)

        # trim hyphens from the beginning and end
        valid_name = name_with_valid_chars.strip("-")
        if not valid_name:
            raise BaseConfigurationError(
                brief=f"failed to create base with hostname {hostname!r}.",
                details="hostname must contain at least one alphanumeric character",
            )

        logger.debug("Using hostname %r", valid_name)
        self._hostname = valid_name

    def _ensure_instance_config_compatible(self, executor: Executor) -> None:
        """Ensure instance configuration is compatible.

        As long as the config is not incompatible (via a mismatched compatibility tag),
        then assume the instance is compatible. This assumption is done because the
        config file may not exist or contain a tag while the set up is in progress.

        :raises BaseCompatibilityError: if instance is incompatible.
        :raises BaseConfigurationError: on other unexpected error.
        """
        try:
            config = InstanceConfiguration.load(
                executor=executor,
                config_path=self._instance_config_path,
            )
        except ValidationError as error:
            raise BaseConfigurationError(
                brief="Failed to parse instance configuration file.",
            ) from error
        # if no config exists, assume base is compatible (likely unfinished setup)
        # XXX: checking the compatibility_tag should be much more strict when called
        # by warmup (warmup will continue if the compatibility tag is missing or none!)
        except FileNotFoundError:
            return

        # make the same assumption as above for empty configs or configs without a tag
        if config is None or config.compatibility_tag is None:
            return

        if config.compatibility_tag != self.compatibility_tag:
            raise BaseCompatibilityError(
                reason=(
                    "Expected image compatibility tag "
                    f"{self.compatibility_tag!r}, found {config.compatibility_tag!r}"
                )
            )
        logger.debug(
            "Instance is compatible with compatibility tag %r", config.compatibility_tag
        )

    def _ensure_setup_completed(self, executor: Executor) -> None:
        """Ensure the instance was fully setup.

        The last step of setting up an instance is to set the `setup` key in the
        instance config file to True. This flag is used to verify that setup was
        completed for the instance.

        :raises BaseCompatibilityError: If setup was not completed.
        """
        try:
            config = InstanceConfiguration.load(
                executor=executor,
                config_path=self._instance_config_path,
            )
        except ValidationError as error:
            raise BaseCompatibilityError(
                reason="failed to parse instance configuration file",
            ) from error
        except FileNotFoundError as error:
            raise BaseCompatibilityError(
                reason="failed to find instance config file",
            ) from error

        if config is None:
            raise BaseCompatibilityError(reason="instance config is empty")

        if not config.setup:
            raise BaseCompatibilityError(reason="instance is marked as not setup")

        logger.debug("Instance has already been setup.")

    def _get_os_release(self, executor: Executor) -> Dict[str, str]:
        """Get the OS release information from an instance's /etc/os-release.

        :returns: Dictionary of key-mappings found in os-release.
        """

        # `lxc exec` will occasionally print an empty string when the CPU is
        # busy. The retry logic here decreases the chance of that.
        def getter(timeout: float) -> str:
            try:
                # Replace encoding errors if it somehow occurs with utf-8. This
                # doesn't need to be perfect for checking compatibility.
                proc = executor.execute_run(
                    command=["cat", "/etc/os-release"],
                    capture_output=True,
                    check=True,
                    text=True,
                    encoding="utf-8",
                    errors="replace",
                    timeout=timeout,
                )
            except subprocess.CalledProcessError as error:
                raise BaseConfigurationError(
                    brief="Failed to read /etc/os-release.",
                    details=details_from_called_process_error(error),
                ) from error
            if not proc.stdout:
                raise BaseConfigurationError(
                    brief="Failed to read /etc/os-release.",
                    details="File appears to be empty.",
                )
            return proc.stdout

        return parse_os_release(
            retry.retry_until_timeout(
                self._timeout_simple or TIMEOUT_SIMPLE,
                self._retry_wait,
                getter,
                error=None,
            )
        )

    @abstractmethod
    def _ensure_os_compatible(self, executor: Executor) -> None:
        """Ensure OS is compatible with Base.

        :raises BaseCompatibilityError: if instance is incompatible.
        :raises BaseConfigurationError: on other unexpected error.
        """

    def get_command_environment(self) -> Dict[str, Optional[str]]:
        """Get command environment to use when executing commands.

        :returns: Dictionary of environment, allowing None as a value to
                  indicate that a value should be unset.
        """
        return self._environment.copy()

    def _update_setup_status(self, executor: Executor, status: bool) -> None:
        """Update the instance config to indicate the status of the setup.

        :param status: True if the setup is complete, False otherwise.
        """
        InstanceConfiguration.update(
            executor=executor,
            data={"setup": status},
            config_path=self._instance_config_path,
        )

    def _update_compatibility_tag(self, executor: Executor) -> None:
        """Update the compatibility_tag in the instance config."""
        InstanceConfiguration.update(
            executor=executor,
            data={"compatibility_tag": self.compatibility_tag},
            config_path=self._instance_config_path,
        )

    def _setup_environment(self, executor: Executor) -> None:
        """Configure /etc/environment.

        If environment is None, reset /etc/environment to the default.
        """
        content = (
            "\n".join(
                [f"{k}={v}" for k, v in self._environment.items() if v is not None]
            )
            + "\n"
        ).encode()

        executor.push_file_io(
            destination=pathlib.PurePosixPath("/etc/environment"),
            content=io.BytesIO(content),
            file_mode="0644",
        )

    def _setup_wait_for_system_ready(self, executor: Executor) -> None:
        """Wait until system is ready."""
        logger.debug("Waiting for environment to be ready...")

        def assert_running(timeout: float) -> None:
            proc = self._execute_run(
                ["systemctl", "is-system-running"],
                executor=executor,
                capture_output=True,
                check=False,
                text=True,
                timeout=timeout,
            )
            system_state = proc.stdout.strip()
            if system_state not in ("running", "degraded"):
                logger.debug("systemctl is-system-running status: %s", system_state)
                raise ValueError

        error = BaseConfigurationError(
            brief="Timed out waiting for environment to be ready."
        )
        retry.retry_until_timeout(
            self._timeout_simple or TIMEOUT_SIMPLE,
            self._retry_wait,
            assert_running,
            error=error,
        )

    def _setup_hostname(self, executor: Executor) -> None:
        """Configure hostname, installing /etc/hostname."""
        executor.push_file_io(
            destination=pathlib.PurePosixPath("/etc/hostname"),
            content=io.BytesIO((self._hostname + "\n").encode()),
            file_mode="0644",
        )

        try:
            self._execute_run(
                ["hostname", "-F", "/etc/hostname"],
                executor=executor,
                timeout=self._timeout_simple,
            )
        except subprocess.CalledProcessError as error:
            raise BaseConfigurationError(
                brief="Failed to set hostname.",
                details=details_from_called_process_error(error),
            ) from error

    def _setup_networkd(self, executor: Executor) -> None:
        """Configure networkd and start it.

        Installs eth0 network configuration using ipv4.
        """
        executor.push_file_io(
            destination=pathlib.PurePosixPath("/etc/systemd/network/10-eth0.network"),
            content=io.BytesIO(
                dedent(
                    """\
                [Match]
                Name=eth0

                [Network]
                DHCP=ipv4
                LinkLocalAddressing=ipv6

                [DHCP]
                RouteMetric=100
                UseMTU=true
                """
                ).encode()
            ),
            file_mode="0644",
        )

        try:
            self._execute_run(
                ["systemctl", "enable", "systemd-networkd"],
                executor=executor,
                timeout=self._timeout_simple,
            )
            self._execute_run(
                ["systemctl", "restart", "systemd-networkd"],
                executor=executor,
                timeout=self._timeout_simple,
            )
        except subprocess.CalledProcessError as error:
            raise BaseConfigurationError(
                brief="Failed to setup systemd-networkd.",
                details=details_from_called_process_error(error),
            ) from error

    def _setup_resolved(self, executor: Executor) -> None:
        """Configure system-resolved to manage resolve.conf."""
        try:
            command = [
                "ln",
                "-sf",
                "/run/systemd/resolve/resolv.conf",
                "/etc/resolv.conf",
            ]
            self._execute_run(command, executor=executor, timeout=self._timeout_simple)

            self._execute_run(
                ["systemctl", "enable", "systemd-resolved"],
                executor=executor,
                timeout=self._timeout_simple,
            )

            self._execute_run(
                ["systemctl", "restart", "systemd-resolved"],
                executor=executor,
                timeout=self._timeout_simple,
            )
        except subprocess.CalledProcessError as error:
            raise BaseConfigurationError(
                brief="Failed to setup systemd-resolved.",
                details=details_from_called_process_error(error),
            ) from error

    def _setup_wait_for_network(self, executor: Executor) -> None:
        """Wait until networking is ready."""
        logger.debug("Waiting for networking to be ready...")
        command = ["getent", "hosts", "snapcraft.io"]

        def check_network(timeout: float) -> None:
            self._execute_run(
                command, executor=executor, check=True, timeout=0.1 * timeout
            )

        error = BaseConfigurationError(
            brief="Timed out waiting for networking to be ready."
        )
        retry.retry_until_timeout(
            self._timeout_simple or math.inf,
            self._retry_wait,
            check_network,
            error=error,
        )

    def _enable_udevd_service(self, executor: Executor) -> None:
        """Enable and start udevd service."""
        try:
            proc = self._execute_run(
                ["systemctl", "is-active", "systemd-udevd"],
                executor=executor,
                capture_output=True,
                check=False,
                text=True,
                timeout=self._timeout_simple,
            )

            state = proc.stdout.strip()
            if state == "active":
                return

            # systemd-udevd is not active, enable and start it
            self._execute_run(
                ["systemctl", "enable", "systemd-udevd"],
                executor=executor,
                timeout=TIMEOUT_SIMPLE,
            )
            self._execute_run(
                ["systemctl", "start", "systemd-udevd"],
                executor=executor,
                timeout=TIMEOUT_SIMPLE,
            )
        except subprocess.CalledProcessError as error:
            raise BaseConfigurationError(
                brief="Failed to enable systemd-udevd service.",
                details=details_from_called_process_error(error),
            ) from error

    def _disable_snapd_cdn(self, executor: Executor) -> None:
        """Disable snapd CDN access if necessary."""
        try:
            # This file is created by launchpad-buildd to stop snapd from
            # using the snap store's CDN when running in Canonical's
            # production build farm, since internet access restrictions may
            # prevent it from doing so but will allow the non-CDN storage
            # endpoint.  If this is in place, then we need to propagate it
            # to containers we create.
            no_cdn = pathlib.Path("/etc/systemd/system/snapd.service.d/no-cdn.conf")
            if no_cdn.exists():
                self._execute_run(
                    ["mkdir", "-p", no_cdn.parent.as_posix()],
                    executor=executor,
                    timeout=TIMEOUT_SIMPLE,
                )

                executor.push_file(source=no_cdn, destination=no_cdn)
        except subprocess.CalledProcessError as error:
            raise BaseConfigurationError(
                brief="Failed to disable snapd CDN.",
                details=details_from_called_process_error(error),
            ) from error

    def _enable_snapd_service(self, executor: Executor) -> None:
        """Create the symlink to /snap and enable the snapd service."""
        try:
            self._execute_run(
                ["ln", "-sf", "/var/lib/snapd/snap", "/snap"],
                executor=executor,
                timeout=TIMEOUT_SIMPLE,
            )

            self._execute_run(
                ["systemctl", "enable", "--now", "snapd.socket"],
                executor=executor,
                timeout=TIMEOUT_SIMPLE,
            )

            # Restart, not start, the service in case the environment
            # has changed and the service is already running.
            self._execute_run(
                ["systemctl", "restart", "snapd.service"],
                executor=executor,
                timeout=TIMEOUT_SIMPLE,
            )
            self._execute_run(
                ["snap", "wait", "system", "seed.loaded"],
                executor=executor,
                timeout=TIMEOUT_SIMPLE,
            )
        except subprocess.CalledProcessError as error:
            raise BaseConfigurationError(
                brief="Failed to enable snapd service.",
                details=details_from_called_process_error(error),
            ) from error

    def _setup_snapd_proxy(self, executor: Executor) -> None:
        """Configure the snapd proxy."""
        try:
            http_proxy = self._environment.get("http_proxy")
            if http_proxy:
                command = ["snap", "set", "system", f"proxy.http={http_proxy}"]
            else:
                command = ["snap", "unset", "system", "proxy.http"]
            self._execute_run(command, executor=executor, timeout=self._timeout_simple)

            https_proxy = self._environment.get("https_proxy")
            if https_proxy:
                command = ["snap", "set", "system", f"proxy.https={https_proxy}"]
            else:
                command = ["snap", "unset", "system", "proxy.https"]
            self._execute_run(command, executor=executor, timeout=self._timeout_simple)

        except subprocess.CalledProcessError as error:
            raise BaseConfigurationError(
                brief="Failed to set the snapd proxy.",
                details=details_from_called_process_error(error),
            ) from error

    def _disable_and_wait_for_snap_refresh(self, executor: Executor) -> None:
        """Disable automatic snap refreshes and wait for refreshes to complete.

        Automatic snap refreshes are disabled because craft-providers manages the
        installation and versions of snaps inside the build environment.

        :param executor: Executor for target container.

        :raises BaseConfigurationError: if snap refreshes cannot be disabled or an
        error occurs while waiting for pending refreshes to complete.
        """
        logger.debug("Holding refreshes for snaps.")

        try:
            executor.execute_run(
                ["snap", "refresh", "--hold"],
                capture_output=True,
                check=True,
                timeout=self._timeout_simple,
            )
        except subprocess.CalledProcessError as error:
            raise BaseConfigurationError(
                brief="Failed to hold snap refreshes.",
                details=details_from_called_process_error(error),
            ) from error

        # a refresh may have started before the hold was set
        logger.debug("Waiting for pending snap refreshes to complete.")
        try:
            executor.execute_run(
                ["snap", "watch", "--last=auto-refresh?"],
                capture_output=True,
                check=True,
                timeout=self._timeout_simple,
            )
        except subprocess.CalledProcessError as error:
            raise BaseConfigurationError(
                brief="Failed to wait for snap refreshes to complete.",
                details=details_from_called_process_error(error),
            ) from error

    def _install_snaps(self, executor: Executor) -> None:
        """Install snaps.

        Snaps will either be installed from the store or injected from the host.
        - If channel is `None` on a linux system, the host snap is injected
          into the provider.
        - If channel is `None` on a non-linux system, an error is raised
          because host injection is not supported on non-linux systems.

        :raises BaseConfigurationError: if the snap cannot be installed
        """
        if not self._snaps:
            logger.debug("No snaps to install.")
            return

        for snap in self._snaps:
            logger.debug(
                "Installing snap %r with channel=%r and classic=%r",
                snap.name,
                snap.channel,
                snap.classic,
            )

            # don't inject snaps on non-linux hosts
            if sys.platform != "linux" and not snap.channel:
                raise BaseConfigurationError(
                    brief=(
                        f"cannot inject snap {snap.name!r} from host on "
                        "a non-linux system"
                    ),
                    resolution=(
                        "install the snap from the store by setting the "
                        "'channel' parameter"
                    ),
                )

            if snap.channel:
                try:
                    snap_installer.install_from_store(
                        executor=executor,
                        snap_name=snap.name,
                        channel=snap.channel,
                        classic=snap.classic,
                    )
                except SnapInstallationError as error:
                    raise BaseConfigurationError(
                        brief=(
                            f"failed to install snap {snap.name!r} from store"
                            f" channel {snap.channel!r} in target environment."
                        ),
                        details=error.details,
                    ) from error
            else:
                try:
                    snap_installer.inject_from_host(
                        executor=executor,
                        snap_name=snap.name,
                        classic=snap.classic,
                    )
                except SnapInstallationError as error:
                    raise BaseConfigurationError(
                        brief=(
                            f"failed to inject host's snap {snap.name!r} "
                            "into target environment."
                        ),
                        details=error.details,
                    ) from error

    def wait_until_ready(self, executor: Executor) -> None:
        """Wait until base instance is ready.

        Ensure minimum-required boot services are running.  This would be used
        when starting an environment's container/VM after already [recently]
        running setup(), e.g. rebooting the instance.  Allows the environment to
        be used without the cost incurred by re-executing the steps
        unnecessarily.

        If timeout is specified, abort operation if time has been exceeded.

        Guarantees provided by this wait:

            - networking available (IP & DNS resolution)

            - system services are started and ready

        :raises ProviderError: on timeout or unexpected error.
        """
        self._setup_wait_for_system_ready(executor=executor)
        self._setup_wait_for_network(executor=executor)

    def _pre_image_check(self, executor: Executor) -> None:
        """Start the setup process and update the status.

        This step usually does not need to be overridden.
        """
        return

    def _image_check(self, executor: Executor) -> None:
        """Check that the image compatibility.

        This step usually does not need to be overridden.
        """
        self._ensure_os_compatible(executor=executor)
        self._ensure_instance_config_compatible(executor=executor)

    def _post_image_check(self, executor: Executor) -> None:
        """Do anything extra image checking.

        This step should be overridden when needed.
        """
        return

    def _pre_setup_os(self, executor: Executor) -> None:
        """Do anything before setting up the OS.

        e.g.
            enable/disable services.
            load kernel modules.

        This step should be overridden when needed.
        """
        return

    def _setup_os(self, executor: Executor) -> None:
        """Set up the OS environment.

        This step should be overridden when needed.
        """
        self._setup_environment(executor=executor)

    def _post_setup_os(self, executor: Executor) -> None:
        """Do anything after setting up the OS.

        e.g.
            additional OS-related configuration.

        This step should be overridden when needed.
        """
        return

    def _pre_setup_network(self, executor: Executor) -> None:
        """Do anything before setting up the basic network.

        e.g.
            enable/disable other network managers.

        This step should be overridden when needed.
        """
        return

    def _setup_network(self, executor: Executor) -> None:
        """Set up the basic network.

        Basic configuration for the systemd and networkd by default.

        Any other configuration requires overriding this step.
        If only part of it needs to be modified you should override only the
        corresponding function.

        This step usually does not need to be overridden.
        """
        self._setup_hostname(executor=executor)

    def _post_setup_network(self, executor: Executor) -> None:
        """Do anything after setting up the basic network.

        e.g.
            additional network-related configuration.

        This step should be overridden when needed.
        """
        return

    def _mount_shared_cache_dirs(self, executor: Executor) -> None:
        """Mount shared cache directories for this base.

        e.g.
            pip cache (normally $HOME/.cache/pip)

        This will only be run if caching is enabled for this instance.

        This step should usually be extended, but may be overridden if common
        cache directories are in unusual locations.
        """
        if self._cache_path is None:
            logger.debug("No cache path set, not mounting cache directories.")
            return

        # Get the real path with additional tags.
        host_base_cache_path = self._cache_path.resolve().joinpath(
            self.compatibility_tag, str(self.alias)
        )

        try:
            host_base_cache_path.mkdir(parents=True, exist_ok=True)
        except OSError as error:
            raise BaseConfigurationError(
                brief=f"Failed to create host cache directory: {host_base_cache_path}"
            ) from error

        guest_cache_proc = executor.execute_run(
            ["bash", "-c", "echo -n ${XDG_CACHE_HOME:-${HOME}/.cache}"],
            capture_output=True,
            text=True,
        )
        guest_base_cache_path = pathlib.Path(guest_cache_proc.stdout)

        # PIP cache
        host_pip_cache_path = host_base_cache_path / "pip"
        host_pip_cache_path.mkdir(parents=True, exist_ok=True)

        guest_pip_cache_path = guest_base_cache_path / "pip"
        executor.execute_run(
            ["mkdir", "-p", guest_pip_cache_path.as_posix()],
        )

        executor.mount(host_source=host_pip_cache_path, target=guest_pip_cache_path)

    def _pre_setup_packages(self, executor: Executor) -> None:
        """Do anything before setting up the packages.

        e.g.
            update package database.

        This step should be overridden when needed.
        """
        return

    @abstractmethod
    def _setup_packages(self, executor: Executor) -> None:
        """Set up the packages.

        This step must be overridden.
        """

    def _post_setup_packages(self, executor: Executor) -> None:
        """Configure the new installed packages.

        This step should be overridden when needed.
        """
        return

    def _pre_setup_snapd(self, executor: Executor) -> None:
        """Do anything before setting up snapd.

        e.g.
            enable depended services.
            check possible incompatibility issues.

        This step should be overridden when needed.
        """
        self._enable_udevd_service(executor=executor)
        self._disable_snapd_cdn(executor=executor)

    @abstractmethod
    def _setup_snapd(self, executor: Executor) -> None:
        """Set up snapd.

        This step must be overridden.
        """

    def _post_setup_snapd(self, executor: Executor) -> None:
        """Configure the new installed snapd.

        This step usually does not need to be overridden.
        """
        self._enable_snapd_service(executor=executor)
        self._disable_and_wait_for_snap_refresh(executor=executor)
        self._setup_snapd_proxy(executor=executor)

    def _warmup_snapd(self, executor: Executor) -> None:
        """Warmup snapd.

        This step usually does not need to be overridden.
        """
        self._setup_snapd_proxy(executor=executor)

    def _pre_setup_snaps(self, executor: Executor) -> None:
        """Do anything before setting up the snaps.

        e.g.
            remove non-snap old packages.

        This step should be overridden when needed.
        """
        return

    def _setup_snaps(self, executor: Executor) -> None:
        """Set up the snaps.

        e.g.
            install snaps.

        This step should be overridden when needed.
        """
        self._install_snaps(executor=executor)

    def _post_setup_snaps(self, executor: Executor) -> None:
        """Configure the new installed snaps.

        This step should be overridden when needed.
        """
        return

    def _pre_clean_up(self, executor: Executor) -> None:
        """Do anything before cleaning up.

        e.g.
            stop affected services.

        This step should be overridden when needed.
        """
        return

    def _clean_up(self, executor: Executor) -> None:
        """Cleanup the OS environment.

        e.g.
            remove unnecessary packages.
            remove cache files.

        This step should be overridden when needed.
        """
        return

    def _post_clean_up(self, executor: Executor) -> None:
        """Do anything needed after cleaning up.

        e.g.
            restart affected services.
            rebuild packages database.

        This step should be overridden when needed.
        """
        return

    def _pre_finish(self, executor: Executor) -> None:
        """Do anything needed before finishing the setup process.

        e.g.
            get the final package / snap versions.

        This step should be overridden when needed.
        """
        return

    @final
    def _finish(self, executor: Executor) -> None:
        """Finish the setup process and update the status.

        This step cannot be overridden.
        """
        self._update_setup_status(executor=executor, status=True)

    @final
    def setup(
        self,
        *,
        executor: Executor,
        timeout: Optional[float] = TIMEOUT_UNPREDICTABLE,
        mount_cache: bool = True,
    ) -> None:
        """Prepare base instance for use by the application.

        Wait for environment to become ready and configure it.  At completion of
        setup, the executor environment should have networking up and have all
        of the installed dependencies required for subsequent use by the
        application.

        Setup should not be called more than once in a given instance to
        refresh/update the environment, use `warmup` for that.

        If timeout is specified, abort operation if time has been exceeded.

        :param executor: Executor for target container.
        :param timeout: Timeout in seconds.
        :param mount_cache: If true, mount the cache directories.

        :raises BaseCompatibilityError: if instance is incompatible.
        :raises BaseConfigurationError: on other unexpected error.
        """
        if timeout is None:
            self._timeout_simple = None
            self._timeout_complex = None
            self._timeout_unpredictable = None
        elif timeout > 0:
            self._timeout_unpredictable = timeout
        else:
            raise BaseConfigurationError(f"Invalid timeout value: {timeout}")

        self._update_setup_status(executor=executor, status=False)

        self._pre_image_check(executor=executor)
        self._image_check(executor=executor)
        self._post_image_check(executor=executor)

        self._update_compatibility_tag(executor=executor)

        if mount_cache:
            self._mount_shared_cache_dirs(executor=executor)

        self._pre_setup_os(executor=executor)
        self._setup_os(executor=executor)
        self._post_setup_os(executor=executor)

        self._setup_wait_for_system_ready(executor=executor)

        self._pre_setup_network(executor=executor)
        self._setup_network(executor=executor)
        self._post_setup_network(executor=executor)

        self._setup_wait_for_network(executor=executor)

        self._pre_setup_packages(executor=executor)
        self._setup_packages(executor=executor)
        self._post_setup_packages(executor=executor)

        self._pre_setup_snapd(executor=executor)
        self._setup_snapd(executor=executor)
        self._post_setup_snapd(executor=executor)

        self._pre_setup_snaps(executor=executor)
        self._setup_snaps(executor=executor)
        self._post_setup_snaps(executor=executor)

        self._pre_clean_up(executor=executor)
        self._clean_up(executor=executor)
        self._post_clean_up(executor=executor)

        self._pre_finish(executor=executor)
        self._finish(executor=executor)

    @final
    def warmup(
        self,
        *,
        executor: Executor,
        timeout: Optional[float] = TIMEOUT_UNPREDICTABLE,
    ) -> None:
        """Prepare a previously created and setup instance for use by the application.

        Ensure the instance is still valid and wait for environment to become ready.

        If timeout is specified, abort operation if time has been exceeded.

        :param executor: Executor for target container.
        :param timeout: Timeout in seconds.

        :raises BaseCompatibilityError: if instance is incompatible.
        :raises BaseConfigurationError: on other unexpected error.
        """
        if timeout is None:
            self._timeout_simple = None
            self._timeout_complex = None
            self._timeout_unpredictable = None
        elif timeout > 0:
            self._timeout_unpredictable = timeout
        else:
            raise BaseConfigurationError(f"Invalid timeout value: {timeout}")

        self._ensure_setup_completed(executor=executor)

        self._pre_image_check(executor=executor)
        self._image_check(executor=executor)
        self._post_image_check(executor=executor)

        self._mount_shared_cache_dirs(executor=executor)

        self._setup_wait_for_system_ready(executor=executor)
        self._setup_wait_for_network(executor=executor)

        self._warmup_snapd(executor=executor)

        self._pre_setup_snaps(executor=executor)
        self._setup_snaps(executor=executor)
        self._post_setup_snaps(executor=executor)

    @staticmethod
    def _network_connected(executor: Executor) -> bool:
        """Check if the network is connected."""
        # bypass the network verification if there is a proxy set for HTTPS
        # (because we're hitting port 443), as bash's TCP functionality will not
        # use it (supporting both lowercase and uppercase names, which is what
        # most applications do)
        if os.getenv("HTTPS_PROXY") or os.getenv("https_proxy"):
            return True

        # check if the port is open using bash's built-in tcp-client, communicating with
        # the HTTPS port on our site
        command = ["bash", "-c", "exec 3<> /dev/tcp/snapcraft.io/443"]
        try:
            # timeout quickly, so it's representative of current state (we don't
            # want for it to hang a lot and then succeed 45 seconds later if network
            # came back); capture the output just for it to not pollute the terminal
            proc = executor.execute_run(
                command, check=False, timeout=10, capture_output=True
            )
        except subprocess.TimeoutExpired:
            return False
        return proc.returncode == 0

    @classmethod
    def _execute_run(
        cls,
        command: List[str],
        *,
        executor: Executor,
        check: bool = True,
        capture_output: bool = True,
        text: bool = False,
        timeout: Optional[float] = None,
        verify_network=False,
    ) -> subprocess.CompletedProcess:
        """Run a command through the executor.

        This is a helper to simplify most common calls and provide extra network
        verification (if indicated) in a central place.

        The default of capture_output is True because it's useful for error reports
        (if the command failed) even if the output is not really wanted as a result
        of the execution.
        """
        if not check and verify_network:
            # if check is False, the caller needs the process result no matter
            # what, it's wrong to also request to verify network, which may
            # raise a different exception
            raise RuntimeError("Invalid check and verify_network combination.")

        try:
            proc = executor.execute_run(
                command,
                check=check,
                capture_output=capture_output,
                text=text,
                timeout=timeout,
            )
        except subprocess.CalledProcessError as exc:
            if verify_network and not cls._network_connected(executor=executor):
                raise NetworkError from exc
            raise
        return proc
